<?php
/**
 * Whups_Driver_sql class - implements a Whups backend for the
 * PEAR::DB abstraction layer.
 *
 * $Horde: whups/lib/Driver/sql.php,v 1.204 2005/01/03 14:35:44 jan Exp $
 *
 * Copyright 2001-2005 Robert E. Coyle <robertecoyle@hotmail.com>
 * Copyright 2001-2005 Chuck Hagenbuch <chuck@horde.org>
 *
 * See the enclosed file LICENSE for license information (BSD). If you
 * did not receive this file, see http://www.horde.org/licenses/bsdl.php.
 *
 * @author  Robert E. Coyle <robertecoyle@hotmail.com>
 * @author  Chuck Hagenbuch <chuck@horde.org>
 * @version $Revision: 1.204 $
 * @package Whups
 */
class Whups_Driver_sql extends Whups_Driver {

    /**
     * The database connection object.
     *
     * @var object DB $_db
     */
    var $_db;

    /**
     * A mapping of attributes from generic Whups names to DB backend
     * fields.
     *
     * @var array $_map
     */
    var $_map = array('id' => 'ticket_id',
                      'summary' => 'ticket_summary',
                      'requester' => 'user_id_requester',
                      'queue' => 'queue_id',
                      'version' => 'version_id',
                      'type' => 'type_id',
                      'state' => 'state_id',
                      'priority' => 'priority_id',
                      'timestamp' => 'ticket_timestamp');

    function addQueue($name, $description)
    {
        // Get a new unique id.
        $new_id = $this->_db->nextId('whups_queues');
        if (is_a($new_id, 'PEAR_Error')) {
            Horde::logMessage($new_id, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $new_id;
        }

        $query = 'INSERT INTO whups_queues (queue_id, queue_name, queue_description) VALUES (' .
                 $this->_db->quote($new_id) . ', ' . $this->_db->quote(String::convertCharset($name, NLS::getCharset(), $this->_params['charset'])) . ', ' . $this->_db->quote(String::convertCharset($description, NLS::getCharset(), $this->_params['charset'])) . ')';
        Horde::logMessage('SQL Query by Whups_Driver_sql::addQueue(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
        $result = $this->_db->query($query);
        if (is_a($result, 'PEAR_Error')) {
            Horde::logMessage($result, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $result;
        }

        return $new_id;
    }

    function addSubject($name)
    {
        // Get a new unique id.
        $new_id = $this->_db->nextId('whups_subjects');
        if (is_a($new_id, 'PEAR_Error')) {
            Horde::logMessage($new_id, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $new_id;
        }

        $query = 'INSERT INTO whups_subjects (subject_id, subject_name) VALUES (' .
                 $this->_db->quote($new_id) . ', ' . $this->_db->quote(String::convertCharset($name, NLS::getCharset(), $this->_params['charset'])) . ')';
        Horde::logMessage('SQL Query by Whups_Driver_sql::addSubject(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
        $result = $this->_db->query($query);
        if (is_a($result, 'PEAR_Error')) {
            Horde::logMessage($result, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $result;
        }

        return $new_id;
    }

    function addType($name, $description)
    {
        // Get a new unique id.
        $new_id = $this->_db->nextId('whups_types');
        if (is_a($new_id, 'PEAR_Error')) {
            Horde::logMessage($new_id, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $new_id;
        }

        $query = 'INSERT INTO whups_types (type_id, type_name, type_description) VALUES (' .
                 $this->_db->quote($new_id) . ', ' . $this->_db->quote(String::convertCharset($name, NLS::getCharset(), $this->_params['charset'])) . ', ' . $this->_db->quote(String::convertCharset($description, NLS::getCharset(), $this->_params['charset'])) . ')';
        Horde::logMessage('SQL Query by Whups_Driver_sql::addType(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
        $result = $this->_db->query($query);
        if (is_a($result, 'PEAR_Error')) {
            Horde::logMessage($result, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $result;
        }

        return $new_id;
    }

    function addState($typeID, $name, $description, $category)
    {
        // Get a new state id.
        $new_id = $this->_db->nextId('whups_states');
        if (is_a($new_id, 'PEAR_Error')) {
            Horde::logMessage($new_id, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $new_id;
        }

        $query = 'INSERT INTO whups_states (state_id, type_id, state_name, state_description, state_category) VALUES (' .
                 $this->_db->quote($new_id) . ', ' . $this->_db->quote($typeID) . ', ' . $this->_db->quote(String::convertCharset($name, NLS::getCharset(), $this->_params['charset'])) . ', ' . $this->_db->quote(String::convertCharset($description, NLS::getCharset(), $this->_params['charset'])) . ', ' . $this->_db->quote(String::convertCharset($category, NLS::getCharset(), $this->_params['charset'])) . ')';
        Horde::logMessage('SQL Query by Whups_Driver_sql::addState(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
        $result = $this->_db->query($query);
        if (is_a($result, 'PEAR_Error')) {
            Horde::logMessage($result, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $result;
        }

        return $new_id;
    }

    function addPriority($typeID, $name, $description)
    {
        // Get a new priority id.
        $new_id = $this->_db->nextId('whups_priorities');
        if (is_a($new_id, 'PEAR_Error')) {
            Horde::logMessage($new_id, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $new_id;
        }

        $query = 'INSERT INTO whups_priorities (priority_id, type_id, priority_name, priority_description) VALUES (' .
                 $this->_db->quote($new_id) . ', ' . $this->_db->quote($typeID) . ', ' . $this->_db->quote(String::convertCharset($name, NLS::getCharset(), $this->_params['charset'])) . ', ' . $this->_db->quote(String::convertCharset($description, NLS::getCharset(), $this->_params['charset'])) . ')';
        Horde::logMessage('SQL Query by Whups_Driver_sql::addPriority(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
        $result = $this->_db->query($query);
        if (is_a($result, 'PEAR_Error')) {
            Horde::logMessage($result, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $result;
        }

        return $new_id;
    }

    function addVersion($queueID, $name, $description)
    {
        // Get a new version id.
        $new_id = $this->_db->nextId('whups_versions');
        if (is_a($new_id, 'PEAR_Error')) {
            Horde::logMessage($new_id, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $new_id;
        }

        $query = 'INSERT INTO whups_versions (version_id, queue_id, version_name, version_description) VALUES (' .
                 $this->_db->quote($new_id) . ', ' . $this->_db->quote($queueID) . ', ' . $this->_db->quote(String::convertCharset($name, NLS::getCharset(), $this->_params['charset'])) . ', ' . $this->_db->quote(String::convertCharset($description, NLS::getCharset(), $this->_params['charset'])) . ')';
        Horde::logMessage('SQL Query by Whups_Driver_sql::addVersion(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
        $result = $this->_db->query($query);
        if (is_a($result, 'PEAR_Error')) {
            Horde::logMessage($result, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $result;
        }

        return $new_id;
    }

    function addTicket($info, $requester)
    {
        $type = $info['type'];
        $state = $info['state'];
        $priority = $info['priority'];
        $queue = $info['queue'];
        $summary = $info['summary'];
        $version = isset($info['version']) ? $info['version'] : null;
        $comment = $info['comment'];
        $attributes = isset($info['attributes']) ? $info['attributes'] : array();

        // Get the new unique ids for this ticket and the initial
        // comment.
        $ticketId = $this->_db->nextId('whups_tickets');
        if (is_a($ticketId, 'PEAR_Error')) {
            Horde::logMessage($ticketId, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $ticketId;
        }

        if (empty($requester)) {
            $requester = $ticketId * -1;
        }

        // Create the ticket.
        $query = 'INSERT INTO whups_tickets (ticket_id, ticket_summary, user_id_requester, type_id, state_id,' .
            " priority_id, queue_id, ticket_timestamp, version_id)" .
            " VALUES ($ticketId, " . $this->_db->quote(String::convertCharset($summary, NLS::getCharset(), $this->_params['charset'])) . ', ' . $this->_db->quote($requester) . ", $type, $state," .
            " $priority, $queue, " . time() . ', ' . $this->_db->quote($version) . ')';
        Horde::logMessage('SQL Query by Whups_Driver_sql::addTicket(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
        $result = $this->_db->query($query);
        if (is_a($result, 'PEAR_Error')) {
            Horde::logMessage($result, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $result;
        }

        if ($requester < 0 && !empty($info['user_email'])) {
            $query = sprintf('INSERT INTO whups_guests (guest_id, guest_email) ' .
                             ' VALUES (%s, %s)',
                             $this->_db->quote((string)$requester),
                             $this->_db->quote($info['user_email']));
            Horde::logMessage('SQL Query by Whups_Driver_sql::addTicket(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
            $result = $this->_db->query($query);
            if (is_a($result, 'PEAR_Error')) {
                Horde::logMessage($result, __FILE__, __LINE__, PEAR_LOG_ERR);
                return $result;
            }
        }

        $commentId = $this->addComment($ticketId, $comment, $requester, isset($info['user_email']) ? $info['user_email'] : null);
        if (is_a($commentId, 'PEAR_Error')) {
            Horde::logMessage($commentId, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $commentId;
        }

        $transaction = $this->updateLog($ticketId, $requester, array('state' => $state,
                                                                     'priority' => $priority,
                                                                     'type' => $type,
                                                                     'summary' => $summary,
                                                                     'comment' => $commentId,
                                                                     'queue' => $queue));
        if (is_a($transaction, 'PEAR_Error')) {
            Horde::logMessage($transaction, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $transaction;
        }

        // Assign the ticket, if requested.
        $owners = array_merge(isset($info['owners']) ? $info['owners'] : array(),
                              isset($info['group_owners']) ? $info['group_owners'] : array());
        foreach ($owners as $owner) {
            $this->addTicketOwner($ticketId, $owner);
            $result = $this->updateLog($ticketId, $requester, array('assign' => $owner), $transaction);
            if (is_a($result, 'PEAR_Error')) {
                Horde::logMessage($result, __FILE__, __LINE__, PEAR_LOG_ERR);
                return $result;
            }
        }

        // Add any supplied attributes for this ticket
        foreach ($attributes as $attribute_id => $attribute_value) {
            $result = $this->_setAttributeValue($ticketId,
                                                $attribute_id,
                                                $attribute_value);
            if (is_a($result, 'PEAR_Error')) {
                Horde::logMessage($result, __FILE__, __LINE__, PEAR_LOG_ERR);
                return $result;
            }
        }

        return $ticketId;
    }

    function addComment($ticket_id, $comment, $creator, $creator_email = null)
    {
        $id = $this->_db->nextId('whups_comments');
        if (is_a($id, 'PEAR_Error')) {
            Horde::logMessage($id, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $id;
        }

        if (empty($creator) || $creator < 0) {
            $creator = '-' . $id . '_comment';
        }

        // Add the row.
        $result = $this->_db->query(sprintf('INSERT INTO whups_comments (comment_id, ticket_id, user_id_creator, comment_text, comment_timestamp) ' .
                                            'VALUES (%d, %d, %s, %s, %d)',
                                            (int)$id,
                                            (int)$ticket_id,
                                            $this->_db->quote($creator),
                                            $this->_db->quote(String::convertCharset($comment, NLS::getCharset(), $this->_params['charset'])),
                                            time()));
        if (is_a($result, 'PEAR_Error')) {
            Horde::logMessage($result, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $result;
        }

        if ($creator < 0 && !empty($creator_email)) {
            $query = sprintf('INSERT INTO whups_guests (guest_id, guest_email) ' .
                             ' VALUES (%s, %s)',
                             $this->_db->quote((string)$creator),
                             $this->_db->quote($creator_email));
            Horde::logMessage('SQL Query by Whups_Driver_sql::addComment(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
            $result = $this->_db->query($query);
            if (is_a($result, 'PEAR_Error')) {
                Horde::logMessage($result, __FILE__, __LINE__, PEAR_LOG_ERR);
                return $result;
            }
        }

        return $id;
    }

    /**
     * Update any details of a ticket that are stored in the main
     * whups_tickets table. Does not update the ticket log (so that it
     * can be used for things low-level enough to not show up
     * there. In general, you should *always* update the log;
     * Whups_Ticket::commit() will take care of this in most cases).
     *
     * @param integer $ticketId    The id of the ticket to update.
     * @param array   $attributes  The array of attributes (key => value) to change.
     *
     * @return boolean | object PEAR_Error  True or an error object.
     */
    function updateTicket($ticketId, $attributes)
    {
        if (!count($attributes)) {
            return true;
        }

        $query = '';
        foreach ($attributes as $field => $value) {
            if (empty($this->_map[$field])) {
                continue;
            }
            $query .= $this->_map[$field] . ' = ' . $this->_db->quote($value) . ', ';
        }

        /* Don't try to execute an empty query (if we didn't find any
         * updates to make). */
        if (empty($query)) {
            return;
        }

        $query = 'UPDATE whups_tickets SET ' . substr($query, 0, -2) . ' WHERE ticket_id = ' . (int)$ticketId;

        Horde::logMessage('SQL Query by Whups_Driver_sql::updateTicket(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
        return $this->_db->query($query);
    }

    function addTicketOwner($ticketId, $owner)
    {
        return $this->_db->query('INSERT INTO whups_ticket_owners (ticket_id, ticket_owner) VALUES (' .
                                 $ticketId . ', ' . $this->_db->quote($owner) . ')');
    }

    function deleteTicketOwner($ticketId, $owner)
    {
        return $this->_db->query('DELETE FROM whups_ticket_owners WHERE ticket_owner = ' . $this->_db->quote($owner) . ' AND ticket_id = ' . $ticketId);
    }

    function deleteTicket($info)
    {
        global $conf;

        $id = (int)$info['id'];

        $tables = array('whups_tickets_listeners',
                        'whups_logs',
                        'whups_comments',
                        'whups_tickets',
                        'whups_attributes');

        if (!empty($conf['vfs']['type'])) {
            require_once 'VFS.php';
            $vfs = &VFS::singleton($conf['vfs']['type'], Horde::getDriverConfig('vfs'));
            if (is_a($vfs, 'PEAR_Error')) {
                return $vfs;
            } else {
                if ($vfs->isFolder(WHUPS_VFS_ATTACH_PATH, $id)) {
                    $result = $vfs->deleteFolder(WHUPS_VFS_ATTACH_PATH, $id, true);
                    if (is_a($result, 'PEAR_Error')) {
                        return $result;
                    }
                }
            }
        }

        // Attempt to clean up everything.
        foreach ($tables as $table) {
            $query = 'DELETE FROM ' . $table . ' WHERE ticket_id = ' . $id;
            Horde::logMessage('SQL Query by Whups_Driver_sql::deleteTicket(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
            $result = $this->_db->query($query);
            if (is_a($result, 'PEAR_Error')) {
                Horde::logMessage($result, __FILE__, __LINE__, PEAR_LOG_ERR);
                return $result;
            }
        }

        return true;
    }

    function executeQuery($query, $vars, $get_details = true, $munge = true)
    {
        $this->tables  = array('whups_tickets' => 1);
        $this->jtables = array();
        $this->joins   = array();

        $where = 'WHERE ' . $query->reduce($this, '_clauseFromQuery', $vars);
        $from  = 'FROM ' . implode(', ', array_keys($this->tables));

        if (count($this->joins) > 0) {
            $joins = implode(' ', $this->joins) . ' ';
        } else {
            $joins = '';
        }

        $sql = "SELECT whups_tickets.ticket_id, 1 $from $joins$where";

        $ids = $this->_db->getAssoc($sql);
        if (is_a($ids, 'PEAR_Error')) {
            Horde::logMessage($ids, __FILE__, __LINE__, PEAR_LOG_ERR);
            $GLOBALS['notification']->push($ids, 'horde.error');
            return array();
        }
        $ids = array_keys($this->_db->getAssoc($sql));

        if (!count($ids)) {
            return array();
        }

        if ($get_details === true) {
            $ids = $this->getTicketsByProperties(array('id' => $ids), $munge);
        }

        return $ids;
    }

    function _clauseFromQuery($args, $type, $ttype, $criterion, $cvalue, $operator, $value)
    {
        switch ($type) {
        case QUERY_TYPE_AND:
            return $this->_concatClauses($args, 'AND');

        case QUERY_TYPE_OR:
            return $this->_concatClauses($args, 'OR');

        case QUERY_TYPE_NOT:
            return $this->_notClause($args);

        case QUERY_TYPE_CRITERION:
            return $this->_criterionClause($ttype, $criterion, $cvalue, $operator, $value);
        }
    }

    function _concatClauses($args, $conjunction)
    {
        $count = count($args);

        if ($count == 0) {
            $result = '';
        } elseif ($count == 1) {
            $result = $args[0];
        } else {
            $result = '(' . $args[0] . ')';

            for ($i = 1; $i < $count; $i++) {
                if ($args[$i] != '') {
                    $result .= ' ' . $conjunction . ' (' . $args[$i] . ')';
                }
            }
        }

        return $result;
    }

    function _notClause($args)
    {
        if (count($args) == 0) {
            return '';
        }

        // TODO: put in a sanity check: count($args) should be 1
        // always.
        return 'NOT (' . $args[0] . ')';
    }

    function _criterionClause($ttype, $criterion, $cvalue, $operator, $value)
    {
        $includeType = false;
        $func        = '';
        $funcend     = '';

        switch ($operator) {
        case OPERATOR_GREATER: $op = '>'; break;
        case OPERATOR_LESS:    $op = '<'; break;
        case OPERATOR_EQUAL:   $op = '='; break;
        case OPERATOR_PATTERN: $op = 'LIKE'; break;

        case OPERATOR_CI_SUBSTRING:
            $value = '%' . $value . '%';
            if ($this->_db->phptype == 'pgsql') {
                $op = 'ILIKE';
            } else {
                $op = 'LIKE';
                $func = 'LOWER(';
                $funcend = ')';
            }
            break;

        case OPERATOR_CS_SUBSTRING:
            // FIXME: Does not work in Postgres.
            $func    = 'LOCATE(' . $this->_db->quote($value) . ', ';
            $funcend = ')';
            $op      = '>';
            $value   = 0;
            break;

        case OPERATOR_WORD:
            // TODO: There might be a better way to avoid missing
            // words at the start and end of the text field.
            $func    = "CONCAT(' ', CONCAT(";
            $funcend = ", ' '))";
            $op      = 'LIKE';
            $value   = "% $value %";
            break;
        }

        $qvalue = $this->_db->quote($value);
        $done = false;
        $text = '';

        switch ($criterion) {
        case CRITERION_ID:        $text = "{$func}whups_tickets.ticket_id{$funcend}"; break;
        case CRITERION_QUEUE:     $text = "{$func}whups_tickets.queue_id{$funcend}"; break;
        case CRITERION_TYPE:      $text = "{$func}whups_tickets.type_id{$funcend}"; break;
        case CRITERION_STATE:     $text = "{$func}whups_tickets.state_id{$funcend}"; $includeType = true; break;
        case CRITERION_PRIORITY:  $text = "{$func}whups_tickets.priority_id{$funcend}"; $includeType = true; break;
        case CRITERION_REQUESTER: $text = "{$func}whups_tickets.user_id_requester{$funcend}"; break;
        case CRITERION_SUMMARY:   $text = "{$func}whups_tickets.ticket_summary{$funcend}"; break;

        case CRITERION_ATTRIBUTE:
            $cvalue = (int)$cvalue;

            if (!isset($this->jtables['whups_attributes'])) {
                $this->jtables['whups_attributes'] = 1;
            }
            $v = $this->jtables['whups_attributes']++;

            $this->joins[] = "LEFT JOIN whups_attributes wa$v ON (whups_tickets.ticket_id = wa$v.ticket_id AND wa$v.attribute_id = $cvalue)";
            $text = "{$func}wa$v.attribute_value{$funcend} $op $qvalue";
            $done = true;
            break;

        case CRITERION_OWNERS:
            if (!isset($this->jtables['whups_ticket_owners'])) {
                $this->jtables['whups_ticket_owners'] = 1;
            }
            $v = $this->jtables['whups_ticket_owners']++;

            $this->joins[] = "LEFT JOIN whups_ticket_owners wto$v ON whups_tickets.ticket_id = wto$v.ticket_id";
            $qvalue = $this->_db->quote('user:' . $value);
            $text = "{$func}wto$v.ticket_owner{$funcend} $op $qvalue";
            $done = true;
            break;

        case CRITERION_GROUPS:
            if (!isset($this->jtables['whups_ticket_owners'])) {
                $this->jtables['whups_ticket_owners'] = 1;
            }
            $v = $this->jtables['whups_ticket_owners']++;

            $this->joins[] = "LEFT JOIN whups_ticket_owners wto$v ON whups_tickets.ticket_id = wto$v.ticket_id";
            $qvalue = $this->_db->quote('group:' . $value);
            $text = "{$func}wto$v.ticket_owner{$funcend} $op $qvalue";
            $done = true;
            break;

        case CRITERION_ADDED_COMMENT:
            if (!isset($this->jtables['whups_comments'])) {
                $this->jtables['whups_comments'] = 1;
            }
            $v = $this->jtables['whups_comments']++;

            $this->joins[] = "LEFT JOIN whups_comments wc$v ON (whups_tickets.ticket_id = wc$v.ticket_id)";
            $text = "{$func}wc$v.user_id_creator{$funcend} $op $qvalue";
            $done = true;
            break;

        case CRITERION_COMMENT:
            if (!isset($this->jtables['whups_comments'])) {
                $this->jtables['whups_comments'] = 1;
            }
            $v = $this->jtables['whups_comments']++;

            $this->joins[] = "LEFT JOIN whups_comments wc$v ON (whups_tickets.ticket_id = wc$v.ticket_id)";
            $text = "{$func}wc$v.comment_text{$funcend} $op $qvalue";
            $done = true;
            break;
        }

        if ($done == false) {
            $text .= " $op $qvalue";
        }

        if ($includeType == true) {
            $text .= "(whups_tickets.type_id = $ttype AND $text)";
        }

        return $text;
    }

    function getTicketsByProperties($info, $munge = true, $perowner = false)
    {
        // Search conditions.
        $where = $this->_generateWhere(
            'whups_tickets',
            array('ticket_id', 'type_id', 'state_id', 'priority_id', 'queue_id'),
            $info, 'integer');

        $where2 = $this->_generateWhere('whups_tickets', array('user_id_requester'), $info, 'string');

        if (empty($where)) {
            $where = $where2;
        } elseif (!empty($where2)) {
            $where .= ' AND ' . $where2;
        }

        // Add summary filter if present.
        if (!empty($info['summary'])) {
            $where = $this->_addWhere($where, 1, 'LOWER(whups_tickets.ticket_summary) LIKE ' . $this->_db->quote('%' . String::lower($info['summary']) . '%'));
        }

        $fields = array('ticket_id as id',
                        'ticket_summary as summary',
                        'user_id_requester',
                        'state_id as state',
                        'type_id as type',
                        'priority_id as priority',
                        'queue_id as queue');

        $fields = $this->_prefixTableToColumns('whups_tickets', $fields);
        $fields .= ', whups_tickets.ticket_timestamp AS timestamp';
        $tables = 'whups_tickets';
        $sums = '';
        $join = '';
        $groupby = 'whups_tickets.ticket_id, whups_tickets.ticket_summary, whups_tickets.user_id_requester, whups_tickets.state_id, whups_tickets.type_id, whups_tickets.priority_id, whups_tickets.queue_id, whups_tickets.ticket_timestamp';

        // State filters.
        if (isset($info['category'])) {
            if (is_array($info['category'])) {
                $cat = '';
                foreach ($info['category'] as $category) {
                    if (!empty($cat)) {
                        $cat .= ' OR ';
                    }
                    $cat .= 'whups_states.state_category = ' . $this->_db->quote($category);
                }
                $cat = ' AND (' . $cat . ')';
            } else {
                $cat = isset($info['category']) ? ' AND whups_states.state_category = ' . $this->_db->quote($info['category']) : '';
            }
        } else {
            $cat = '';
        }

        // Type filters.
        if (isset($info['type_id'])) {
            if (is_array($info['type_id'])) {
                $t = array();
                foreach ($info['type_id'] as $type) {
                    $t[] = 'whups_tickets.type_id = ' . $this->_db->quote($type);
                }
                $t = ' AND (' . implode(' OR ', $t) . ')';
            } else {
                $t = isset($info['type_id']) ? ' AND whups_tickets.type_id = ' . $this->_db->quote($info['type_id']) : '';
            }

            $this->_addWhere($where, $t, $t);
        }

        $nouc = isset($info['nouc']) ? " AND whups_states.state_category <> 'unconfirmed'" : '';
        $nores = isset($info['nores']) ? " AND whups_states.state_category <> 'resolved'" : '';
        $nonew = isset($info['nonew']) ? " AND whups_states.state_category <> 'new'" : '';
        $noass = isset($info['noass']) ? " AND whups_states.state_category <> 'assigned'" : '';

        $uc = isset($info['uc']) ? " AND whups_states.state_category = 'unconfirmed'" : '';
        $res = isset($info['res']) ? " AND whups_states.state_category = 'resolved'" : '';
        $new = isset($info['new']) ? " AND whups_states.state_category = 'new'" : '';
        $ass = isset($info['ass']) ? " AND whups_states.state_category = 'assigned'" : '';

        // If there are any state filters, add them in.
        if ($nouc || $nores || $nonew || $noass ||
            $uc || $res || $new || $ass || $cat) {
            $where = $this->_addWhere($where, 1, "(whups_tickets.type_id = whups_states.type_id AND whups_tickets.state_id = whups_states.state_id$nouc$nores$nonew$noass$uc$res$new$ass$cat)");
        }

        // Initialize join clauses.
        $join = '';

        // Handle owner properties.
        if (isset($info['owner'])) {
            $join .= ' INNER JOIN whups_ticket_owners ON whups_tickets.ticket_id = whups_ticket_owners.ticket_id AND ';
            if (is_array($info['owner'])) {
                $clauses = array();
                foreach ($info['owner'] as $owner) {
                    $clauses[] = 'whups_ticket_owners.ticket_owner = ' . $this->_db->quote($owner);
                }
                $join .= '(' . implode(' OR ', $clauses) . ')';
            } else {
                $join .= 'whups_ticket_owners.ticket_owner = ' . $this->_db->quote($info['owner']);
            }
        }
        if (isset($info['notowner'])) {
            $join .= ' LEFT JOIN whups_ticket_owners ON whups_tickets.ticket_id = whups_ticket_owners.ticket_id AND whups_ticket_owners.ticket_owner = ' . $this->_db->quote($info['notowner']);
            $where = $this->_addWhere($where, 1, 'whups_ticket_owners.ticket_id IS NULL');
        }

        if ($munge) {
            $fields = "$fields, " .
                'whups_types.type_name AS type_name, ' .
                'whups_states.state_name AS state_name, ' .
                'whups_states.state_category AS state_category, '.
                'whups_queues.queue_name AS queue_name, ' .
                'whups_priorities.priority_name AS priority_name, ' .
                'whups_versions.version_name AS version_name, ' .
                'whups_versions.version_description AS version_description';

            $join .= ' INNER JOIN whups_types ON whups_tickets.type_id = whups_types.type_id';
            $join .= ' INNER JOIN whups_states ON whups_tickets.state_id = whups_states.state_id';
            $join .= ' INNER JOIN whups_queues ON whups_tickets.queue_id = whups_queues.queue_id';
            $join .= ' INNER JOIN whups_priorities ON whups_tickets.priority_id = whups_priorities.priority_id';
            $join .= ' INNER JOIN whups_states state2 ON whups_tickets.type_id = state2.type_id';
            $join .= ' LEFT JOIN whups_versions ON whups_tickets.version_id = whups_versions.version_id';
            $join .= ' LEFT JOIN whups_logs w1 ON w1.ticket_id = whups_tickets.ticket_id AND w1.log_type = \'state\' AND w1.log_value_num = state2.state_id AND state2.state_category = \'assigned\'';
            $join .= ' LEFT JOIN whups_logs w2 ON whups_states.state_category = \'resolved\' AND w2.ticket_id = whups_tickets.ticket_id AND w2.log_type = \'state\' AND w2.log_value_num = state2.state_id AND state2.state_category = \'resolved\'';

            $sums = ', MAX(w1.log_timestamp) AS date_assigned, MAX(w2.log_timestamp) AS date_resolved';

            $groupby .= ', whups_types.type_name, whups_states.state_name, whups_states.state_category, whups_queues.queue_name, whups_priorities.priority_name, whups_versions.version_name, whups_versions.version_description';
        }

        if ($perowner) {
            $join .= ' LEFT JOIN whups_ticket_owners ON whups_tickets.ticket_id = whups_ticket_owners.ticket_id';
            $fields .= ', whups_ticket_owners.ticket_owner AS owner';
            $groupby .= ', whups_ticket_owners.ticket_owner';
        }

        $query = "SELECT $fields$sums FROM $tables$join " . (!empty($where) ? "WHERE $where " : '') . 'GROUP BY ' . $groupby . ' ORDER BY whups_tickets.priority_id, whups_tickets.ticket_timestamp';
        Horde::logMessage('SQL Query by Whups_Driver_sql::getTicketsByProperties(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);

        $info = $this->_db->getAll($query, null, DB_FETCHMODE_ASSOC);
        if (is_a($info, 'PEAR_Error')) {
            Horde::logMessage($info, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $info;
        }

        if (!count($info)) {
            return array();
        }

        $info = String::convertCharset($info, $this->_params['charset']);

        $tickets = array();
        foreach ($info as $ticket) {
            $tickets[$ticket['id']] = $ticket;
        }

        $owners = $this->getOwners(array_keys($tickets));
        if (is_a($owners, 'PEAR_Error')) {
            return $owners;
        }

        foreach ($owners as $row) {
            if (empty($tickets[$row['id']]['owners'])) {
                $tickets[$row['id']]['owners'] = array();
            }
            $tickets[$row['id']]['owners'][] = $row['owner'];
        }

        return array_values($tickets);
    }

    function getTicketDetails($ticket)
    {
        global $conf;
        $info = array('id' => $ticket);

        $result = $this->getTicketsByProperties($info);
        if (is_a($result, 'PEAR_Error')) {
            Horde::logMessage($result, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $result;
        } elseif (!isset($result[0])) {
            return PEAR::raiseError(sprintf(_("Ticket %s was not found."), $ticket));
        } elseif (!in_array($result[0]['queue'],
                            array_flip(Whups::permissionsFilter($this->getQueues(), 'queue', PERMS_READ)))) {
            return PEAR::raiseError(sprintf(_("You do not have permission to access this ticket (%s)."), $ticket),
                                    0);
        }

        return $result[0];
    }

    function getGuestEmail($guest_id)
    {
        static $guestCache;

        if (!isset($guestCache[$guest_id])) {
            $query = 'SELECT guest_email FROM whups_guests WHERE guest_id = ' . $this->_db->quote($guest_id);
            Horde::logMessage('SQL Query by Whups_Driver_sql::getGuestEmail(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
            $result = $this->_db->getOne($query);
            if (is_a($result, 'PEAR_Error')) {
                return $result;
            }
            $guestCache[$guest_id] = $result;
        }
        return $guestCache[$guest_id];
    }

    function _getHistory($ticket)
    {
        $where = 'whups_logs.ticket_id = ' . (int)$ticket;
        $join  = 'LEFT JOIN whups_comments ON whups_logs.log_type = \'comment\' AND whups_logs.log_value_num = whups_comments.comment_id ';
        $join .= 'LEFT JOIN whups_states ON whups_logs.log_type = \'state\' AND whups_logs.log_value_num = whups_states.state_id ';
        $join .= 'LEFT JOIN whups_priorities ON whups_logs.log_type = \'priority\' AND whups_logs.log_value_num = whups_priorities.priority_id ';
        $join .= 'LEFT JOIN whups_types ON whups_logs.log_type = \'type\' AND whups_logs.log_value_num = whups_types.type_id ';
        $join .= 'LEFT JOIN whups_queues ON whups_logs.log_type = \'queue\' AND whups_logs.log_value_num = whups_queues.queue_id ';

        $fields = $this->_prefixTableToColumns('whups_comments', array('comment_text'));
        $fields .= ', whups_logs.log_timestamp AS timestamp, whups_logs.ticket_id';
        $fields .= ', whups_logs.log_type, whups_logs.log_value, whups_logs.log_id, whups_logs.transaction_id, whups_logs.user_id';
        $fields .= ', whups_priorities.priority_name, whups_states.state_name, whups_types.type_name, whups_queues.queue_name';

        $query = "SELECT $fields FROM whups_logs $join WHERE $where ORDER BY whups_logs.transaction_id";
        Horde::logMessage('SQL Query by Whups_Driver_sql::_getHistory(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);

        $history = $this->_db->getAll($query, null, DB_FETCHMODE_ASSOC);
        if (is_a($history, 'PEAR_Error')) {
            Horde::logMessage($history, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $history;
        }

        return String::convertCharset($history, $this->_params['charset']);
    }

    function getAllQueueInfo()
    {
        $query = 'SELECT queue_id, queue_name, queue_description, queue_subjectlist, queue_versioned FROM whups_queue ORDER BY queue_id';
        Horde::logMessage('SQL Query by Whups_Driver_sql::getAllQueueInfo(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);

        $info = $this->_db->getAll($query, null, DB_FETCHMODE_ASSOC);
        if (is_a($info, 'PEAR_Error')) {
            Horde::logMessage($info, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $info;
        }

        return String::convertCharset($info, $this->_params['charset']);
    }

    function getQueueName($queueID)
    {
        $query = sprintf('SELECT queue_name
                          FROM whups_queues
                          WHERE queue_id = %s', $queueID);
        Horde::logMessage('SQL Query by Whups_Driver_sql::getQueueName(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);

        $name = $this->_db->getOne($query);
        if (is_a($name, 'PEAR_Error')) {
            Horde::logMessage($name, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $name;
        }

        return String::convertCharset($name, $this->_params['charset']);
    }

    function getQueue($queueID)
    {
        $query = 'SELECT queue_id, queue_name, queue_description, queue_subjectlist, queue_versioned FROM whups_queues WHERE queue_id = ' . (int)$queueID;
        Horde::logMessage('SQL Query by Whups_Driver_sql::getQueue(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
        $queue = $this->_db->getAssoc($query);
        if (is_a($queue, 'PEAR_Error')) {
            Horde::logMessage($queue, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $queue;
        } elseif (!count($queue)) {
            return $queue;
        }

        $queue = String::convertCharset($queue, $this->_params['charset']);
        return array('id' => $queueID,
                     'name' => $queue[$queueID][0],
                     'description' => $queue[$queueID][1],
                     'subjectlist' => $queue[$queueID][2],
                     'versioned' => $queue[$queueID][3]);
    }

    function getQueues()
    {
        $query = 'SELECT queue_id, queue_name FROM whups_queues ORDER BY queue_name';
        Horde::logMessage('SQL Query by Whups_Driver_sql::getQueues(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
        $queues = $this->_db->getAssoc($query);
        if (is_a($queues, 'PEAR_Error')) {
            Horde::logMessage($queues, __FILE__, __LINE__, PEAR_LOG_ERR);
            return array();
        }

        return String::convertCharset($queues, $this->_params['charset']);
    }

    function updateQueue($queueID, $name, $description, $types, $subjectlist, $subjects, $versioned)
    {
        // First update the queue entry itself.
        $query = 'UPDATE whups_queues SET' .
                 ' queue_name = ' . $this->_db->quote(String::convertCharset($name, NLS::getCharset(), $this->_params['charset'])) . ',' .
                 ' queue_description = ' . $this->_db->quote(String::convertCharset($description, NLS::getCharset(), $this->_params['charset'])) . ',' .
                 ' queue_subjectlist = ' . (is_null($subjectlist) ? 0 : 1) . ',' .
                 ' queue_versioned = ' . (is_null($versioned) ? 0 : 1) .
                 ' WHERE queue_id = ' . $this->_db->quote($queueID);
        Horde::logMessage('SQL Query by Whups_Driver_sql::updateQueue(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
        $result = $this->_db->query($query);
        if (is_a($result, 'PEAR_Error')) {
            Horde::logMessage($result, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $result;
        }

        // Clear all previous type-queue associations.
        $query = 'DELETE FROM whups_types_queues WHERE queue_id = ' . $this->_db->quote($queueID);
        Horde::logMessage('SQL Query by Whups_Driver_sql::updateQueue(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
        $result = $this->_db->query($query);
        if (is_a($result, 'PEAR_Error')) {
            Horde::logMessage($result, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $result;
        }

        // Add the new associations.
        if (is_array($types)) {
            foreach ($types as $typeID) {
                $query = 'INSERT INTO whups_types_queues (queue_id, type_id) VALUES (' .
                         $this->_db->quote($queueID) . ', ' . $this->_db->quote($typeID) . ')';
                Horde::logMessage('SQL Query by Whups_Driver_sql::updateQueue(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
                $result = $this->_db->query($query);
                if (is_a($result, 'PEAR_Error')) {
                    Horde::logMessage($result, __FILE__, __LINE__, PEAR_LOG_ERR);
                    return $result;
                }
            }
        }

        // Clear all previous subject-queue associations.
        $query = 'DELETE FROM whups_queues_subjects WHERE queue_id = ' . $this->_db->quote($queueID);
        Horde::logMessage('SQL Query by Whups_Driver_sql::updateQueue(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
        $result = $this->_db->query($query);
        if (is_a($result, 'PEAR_Error')) {
            Horde::logMessage($result, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $result;
        }

        // add the new associations
        if (is_array($subjects)) {
            foreach ($subjects as $subjectID) {
                $query = 'INSERT INTO whups_queues_subjects (queue_id, subject_id) VALUES (' .
                         $this->_db->quote($queueID) . ', ' . $this->_db->quote($subjectID) . ')';
                Horde::logMessage('SQL Query by Whups_Driver_sql::updateQueue(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
                $result = $this->_db->query($query);
                if (is_a($result, 'PEAR_Error')) {
                    Horde::logMessage($result, __FILE__, __LINE__, PEAR_LOG_ERR);
                    return $result;
                }
            }
        }

        return true;
    }

    function deleteQueue($queueID)
    {
        $tables = array('whups_queues_users',
                        'whups_queues_subjects',
                        'whups_types_queues',
                        'whups_versions',
                        'whups_queues');
        foreach ($tables as $table) {
            $query = 'DELETE FROM ' . $table . ' WHERE queue_id = ' . $this->_db->quote($queueID);
            Horde::logMessage('SQL Query by Whups_Driver_sql::deleteQueue(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
            $result = $this->_db->query($query);
            if (is_a($result, 'PEAR_Error')) {
                Horde::logMessage($result, __FILE__, __LINE__, PEAR_LOG_ERR);
                return $result;
            }
        }

        return true;
    }

    function updateTypesQueues($tmPairs)
    {
        // Do this as a transaction.
        $this->_db->autoCommit(false);

        // Delete existing associations.
        $query = 'DELETE FROM whups_types_queues';
        Horde::logMessage('SQL Query by Whups_Driver_sql::updateTypesQueues(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
        $result = $this->_db->query($query);
        if (is_a($result, 'PEAR_Error')) {
            Horde::logMessage($result, __FILE__, __LINE__, PEAR_LOG_ERR);
            $this->_db->rollback();
            $this->_db->autoCommit(true);
            return $result;
        }

        // Insert new associations.
        foreach ($tmPairs as $pair) {
            $query = 'INSERT INTO whups_types_queues (queue_id, type_id) VALUES (' .
                (int)$pair[0] . ', ' . (int)$pair[1] . ')';
            Horde::logMessage('SQL Query by Whups_Driver_sql::updateTypesQueues(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
            $result = $this->_db->query($query);
            if (is_a($result, 'PEAR_Error')) {
                Horde::logMessage($result, __FILE__, __LINE__, PEAR_LOG_ERR);
                $this->_db->rollback();
                $this->_db->autoCommit(true);
                return $result;
            }
        }

        $result = $this->_db->commit();
        if (is_a($result, 'PEAR_Error')) {
            Horde::logMessage($result, __FILE__, __LINE__, PEAR_LOG_ERR);
            $this->_db->rollback();
            $this->_db->autoCommit(true);
            return $result;
        }

        $this->_db->autoCommit(true);
    }

    function getSubject($subjectID)
    {
        if (empty($subjectID)) {
            return false;
        }
        $query = 'SELECT subject_name FROM whups_subjects WHERE subject_id = ' . $this->_db->quote($subjectID);
        Horde::logMessage('SQL Query by Whups_Driver_sql::getSubject(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
        $subject = $this->_db->getOne($query);
        if (is_a($subject, 'PEAR_Error')) {
            Horde::logMessage($subject, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $subject;
        }

        return array('id' => $subjectID,
                     'name' => String::convertCharset($subject, $this->_params['charset']));
    }

    function getSubjects()
    {
        $query = 'SELECT subject_id, subject_name FROM whups_subjects ORDER BY subject_name';
        Horde::logMessage('SQL Query by Whups_Driver_sql::getSubjects(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);

        $subjects = $this->_db->getAssoc($query);
        if (is_a($subjects, 'PEAR_Error')) {
            Horde::logMessage($subjects, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $subjects;
        }

        return String::convertCharset($subjects, $this->_params['charset']);
    }

    function getQueueSubjects($queueID)
    {
        $query = 'SELECT s.subject_id, s.subject_name FROM whups_queues_subjects sm, whups_subjects s' .
                 ' WHERE sm.queue_id = ' . $this->_db->quote($queueID) . ' AND sm.subject_id = s.subject_id ORDER BY s.subject_name';
        Horde::logMessage('SQL Query by Whups_Driver_sql::getQueueSubjects(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);

        $subjects = $this->_db->getAssoc($query);
        if (is_a($subjects, 'PEAR_Error')) {
            Horde::logMessage($subjects, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $subjects;
        }

        return String::convertCharset($subjects, $this->_params['charset']);
    }

    function getQueueSubjectNames($queueID)
    {
        $query = 'SELECT s.subject_name, s.subject_name FROM whups_queues_subjects sm, whups_subjects s' .
                 ' WHERE sm.queue_id = ' . $this->_db->quote($queueID) . ' AND sm.subject_id = s.subject_id ORDER BY s.subject_name';
        Horde::logMessage('SQL Query by Whups_Driver_sql::getQueueSubjectNames(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);

        $names = $this->_db->getAssoc($query);
        if (is_a($names, 'PEAR_Error')) {
            Horde::logMessage($names, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $names;
        }

        return String::convertCharset($names, $this->_params['charset']);
    }

    function updateSubject($subjectID, $name)
    {
        // First update the subject entry itself.
        $query = 'UPDATE whups_subjects SET' .
                 ' subject_name = ' . $this->_db->quote(String::convertCharset($name, NLS::getCharset(), $this->_params['charset'])) .
                 ' WHERE subject_id = ' . $this->_db->quote($subjectID);

        Horde::logMessage('SQL Query by Whups_Driver_sql::updateSubject(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
        $result = $this->_db->query($query);
        if (is_a($result, 'PEAR_Error')) {
            Horde::logMessage($result, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $result;
        }

        return true;
    }

    function deleteSubject($subjectID)
    {
        $query = 'DELETE FROM whups_subjects WHERE subject_id = ' . $this->_db->quote($subjectID);
        Horde::logMessage('SQL Query by Whups_Driver_sql::deleteSubject(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
        return $this->_db->query($query);
    }

    function getQueueUsers($queueID)
    {
        $query = 'SELECT user_uid AS u1, user_uid AS u2 FROM whups_queues_users' .
                 ' WHERE queue_id = ' . $this->_db->quote($queueID) . ' ORDER BY u1';
        Horde::logMessage('SQL Query by Whups_Driver_sql::getQueueUsers(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
        $users = $this->_db->getAssoc($query);
        if (is_a($users, 'PEAR_Error')) {
            Horde::logMessage($users, __FILE__, __LINE__, PEAR_LOG_ERR);
            return array();
        }

        return $users;
    }

    function addQueueUser($queueID, $userID)
    {
        if (!is_array($userID)) {
            $userID = array($userID);
        }
        foreach ($userID as $user) {
            $query = 'INSERT INTO whups_queues_users (queue_id, user_uid) values (' .
                     $this->_db->quote($queueID) . ', ' . $this->_db->quote($user) . ')';
            Horde::logMessage('SQL Query by Whups_Driver_sql::addQueueUser(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
            $result = $this->_db->query($query);
            if (is_a($result, 'PEAR_Error')) {
                Horde::logMessage($result, __FILE__, __LINE__, PEAR_LOG_ERR);
                return $result;
            }
        }
        return true;
    }

    function removeQueueUser($queueID, $userID)
    {
        $query = 'DELETE FROM whups_queues_users' .
                 ' WHERE queue_id = ' . $this->_db->quote($queueID) .
                 ' AND user_uid = ' . $this->_db->quote($userID);
        Horde::logMessage('SQL Query by Whups_Driver_sql::removeQueueUser(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
        return $this->_db->query($query);
    }

    function updateQueueUsers($queueID, $users)
    {
        // clear all previous type-queue associations
        $query = 'DELETE FROM whups_mosules_users WHERE queue_id = ' . $this->_db->quote($queueID);
        Horde::logMessage('SQL Query by Whups_Driver_sql::updateQueueUser(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
        $result = $this->_db->query($query);
        if (is_a($result, 'PEAR_Error')) {
            Horde::logMessage($result, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $result;
        }

        // add the new associations
        foreach ($types as $typeID) {
            $query = 'INSERT INTO whups_queues_users (queue_id, type_id) VALUES (' .
                     $this->_db->quote($queueID) . ', ' . $this->_db->quote($typeID) . ')';
            Horde::logMessage('SQL Query by Whups_Driver_sql::updateQueueUser(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
            $result = $this->_db->query($query);
            if (is_a($result, 'PEAR_Error')) {
                Horde::logMessage($result, __FILE__, __LINE__, PEAR_LOG_ERR);
                return $result;
            }
        }

        // whew, everything worked
        return true;
    }

    function getType($typeID)
    {
        if (empty($typeID)) {
            return false;
        }
        $query = 'SELECT type_id, type_name, type_description FROM whups_types WHERE type_id = ' . $this->_db->quote($typeID);
        Horde::logMessage('SQL Query by Whups_Driver_sql::getType(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);

        $type = $this->_db->getAssoc($query);
        if (is_a($type, 'PEAR_Error')) {
            Horde::logMessage($type, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $type;
        }

        $type = String::convertCharset($type, $this->_params['charset']);
        return array('id' => $typeID,
                     'name' => isset($type[$typeID][0]) ? $type[$typeID][0] : '',
                     'description' => isset($type[$typeID][1]) ? $type[$typeID][1] : '');
    }

    function getTypes($queueID)
    {
        $query = 'SELECT t.type_id, t.type_name FROM whups_types t, whups_types_queues tm WHERE tm.queue_id = ' . $this->_db->quote($queueID) . ' AND tm.type_id = t.type_id ORDER BY t.type_name';
        Horde::logMessage('SQL Query by Whups_Driver_sql::getTypes(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
        $types = $this->_db->getAssoc($query);
        if (is_a($types, 'PEAR_Error')) {
            Horde::logMessage($types, __FILE__, __LINE__, PEAR_LOG_ERR);
            return array();
        }

        return String::convertCharset($types, $this->_params['charset']);
    }

    function getTypeIds($queueID)
    {
        $query = 'SELECT type_id FROM whups_types_queues WHERE queue_id = ' . $this->_db->quote($queueID) . ' ORDER BY type_id';
        Horde::logMessage('SQL Query by Whups_Driver_sql::getTypeIds(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
        return $this->_db->getAll($query, null, DB_FETCHMODE_ASSOC);
    }

    function getAllTypes()
    {
        $query = 'SELECT type_id, type_name FROM whups_types ORDER BY type_name';
        Horde::logMessage('SQL Query by Whups_Driver_sql::getAllTypes(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
        $types = $this->_db->getAssoc($query);
        if (is_a($types, 'PEAR_Error')) {
            Horde::logMessage($types, __FILE__, __LINE__, PEAR_LOG_ERR);
            return array();
        }

        return String::convertCharset($types, $this->_params['charset']);
    }

    function getAllTypeInfo()
    {
        $query = 'SELECT type_id, type_name, type_description FROM whups_types ORDER BY type_id';
        Horde::logMessage('SQL Query by Whups_Driver_sql::getAllTypeInfo(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);

        $info = $this->_db->getAll($query, null, DB_FETCHMODE_ASSOC);
        if (is_a($info, 'PEAR_Error')) {
            Horde::logMessage($info, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $info;
        }

        return String::convertCharset($info, $this->_params['charset']);
    }

    function getTypeName($type)
    {
        $query = "SELECT type_name FROM whups_types WHERE type_id = $type";
        Horde::logMessage('SQL Query by Whups_Driver_sql::getTypeName(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);

        $name = $this->_db->getOne($query);
        if (is_a($name, 'PEAR_Error')) {
            Horde::logMessage($name, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $name;
        }

        return String::convertCharset($name, $this->_params['charset']);
    }

    function updateType($typeID, $name, $description)
    {
        $query = 'UPDATE whups_types SET' .
                 ' type_name = ' . $this->_db->quote(String::convertCharset($name, NLS::getCharset(), $this->_params['charset'])) . ',' .
                 ' type_description = ' . $this->_db->quote(String::convertCharset($description, NLS::getCharset(), $this->_params['charset'])) .
                 ' WHERE type_id = ' . $this->_db->quote($typeID);
        Horde::logMessage('SQL Query by Whups_Driver_sql::updateType(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
        return $this->_db->query($query);
    }

    function deleteType($typeID)
    {
        $this->_db->query('DELETE FROM whups_states WHERE type_id = ' . (int)$typeID);
        $this->_db->query('DELETE FROM whups_priorities WHERE type_id = ' . (int)$typeID);
        $this->_db->query('DELETE FROM whups_attributes_desc WHERE type_id = ' . (int)$typeID);

        $query = 'DELETE FROM whups_types WHERE type_id = ' . (int)$typeID;
        Horde::logMessage('SQL Query by Whups_Driver_sql::deleteType(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
        return $this->_db->query($query);
    }

    function getStates($type, $category = '', $notcategory = '')
    {
        if (empty($type)) {
            return array();
        }

        $where = "type_id = $type";

        if (!is_array($category)) {
            $where = $this->_addWhere($where, $category, 'state_category = ' . $this->_db->quote($category));
        } else {
            $clauses = array();
            foreach ($category as $cat) {
                $clauses[] = 'state_category = ' . $this->_db->quote($cat);
            }
            if (count($clauses))
                $where = $this->_addWhere($where, $cat, implode(' OR ', $clauses));
        }

        if (!is_array($notcategory)) {
            $where = $this->_addWhere($where, $notcategory, 'state_category <> ' . $this->_db->quote($notcategory));
        } else {
            $clauses = array();
            foreach ($notcategory as $notcat) {
                $clauses[] = 'state_category <> ' . $this->_db->quote($notcat);
            }
            if (count($clauses)) {
                $where = $this->_addWhere($where, $notcat, implode(' OR ', $clauses));
            }
        }

        $query = "SELECT state_id, state_name FROM whups_states WHERE $where ORDER BY state_category, state_name";
        Horde::logMessage('SQL Query by Whups_Driver_sql::getStates(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);

        $states = $this->_db->getAssoc($query);
        if (is_a($states, 'PEAR_Error')) {
            Horde::logMessage($states, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $states;
        }

        return String::convertCharset($states, $this->_params['charset']);
    }

    function getStateName($type, $state)
    {
        $query = "SELECT state_name FROM whups_states WHERE type_id = $type AND state_id = $state";
        Horde::logMessage('SQL Query by Whups_Driver_sql::getStateName(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);

        $name = $this->_db->getOne($query);
        if (is_a($name, 'PEAR_Error')) {
            Horde::logMessage($name, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $name;
        }

        return String::convertCharset($name, $this->_params['charset']);
    }

    function getState($stateID)
    {
        if (empty($stateID)) {
            return false;
        }
        $query = 'SELECT state_id, state_name, state_description, state_category FROM whups_states WHERE state_id = ' . $this->_db->quote($stateID);
        Horde::logMessage('SQL Query by Whups_Driver_sql::getState(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
        $state = $this->_db->getAssoc($query);
        if (is_a($state, 'PEAR_Error')) {
            Horde::logMessage($state, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $state;
        }

        $state = String::convertCharset($state, $this->_params['charset']);
        return array('id' => $stateID,
                     'name' => isset($state[$stateID][0]) ? $state[$stateID][0] : '',
                     'description' => isset($state[$stateID][1]) ? $state[$stateID][1] : '',
                     'category' => isset($state[$stateID][2]) ? $state[$stateID][2] : '');
    }

    function getAllStateInfo($type)
    {
        $type = $this->_db->quote($type);
        $query = 'SELECT state_id, state_name, state_description, state_category FROM whups_states ' .
                 "WHERE type_id = $type " .
                 'ORDER BY state_id';
        Horde::logMessage('SQL Query by Whups_Driver_sql::getAllStateInfo(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);

        $info = $this->_db->getAll($query, null, DB_FETCHMODE_ASSOC);
        if (is_a($info, 'PEAR_Error')) {
            Horde::logMessage($info, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $info;
        }

        return String::convertCharset($info, $this->_params['charset']);
    }

    function updateState($stateID, $name, $description, $category)
    {
        $query = 'UPDATE whups_states SET' .
                 ' state_name = ' . $this->_db->quote(String::convertCharset($name, NLS::getCharset(), $this->_params['charset'])) . ',' .
                 ' state_description = ' . $this->_db->quote(String::convertCharset($description, NLS::getCharset(), $this->_params['charset'])) . ',' .
                 ' state_category = ' . $this->_db->quote(String::convertCharset($category, NLS::getCharset(), $this->_params['charset'])) .
                 ' WHERE state_id = ' . $this->_db->quote($stateID);
        Horde::logMessage('SQL Query by Whups_Driver_sql::updateState(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
        return $this->_db->query($query);
    }

    function deleteState($state_id)
    {
        $query = 'DELETE FROM whups_states WHERE state_id = ' . (int)$state_id;
        Horde::logMessage('SQL Query by Whups_Driver_sql::deleteState(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
        return $this->_db->query($query);
    }

    function isCategory($category, $state_id)
    {
        $query = 'SELECT 1 FROM whups_states WHERE state_id = ' . (int)$state_id . ' AND state_category = ' . $this->_db->quote($category);
        Horde::logMessage('SQL Query by Whups_Driver_sql::isCategory(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
        return $this->_db->getOne($query);
    }

    function getAllPriorityInfo($type)
    {
        $type = (int)$type;
        $query = "SELECT priority_id, priority_name, priority_description FROM whups_priorities WHERE type_id = $type ORDER BY priority_id";
        Horde::logMessage('SQL Query by Whups_Driver_sql::getAllPriorityInfo(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);

        $info = $this->_db->getAll($query, null, DB_FETCHMODE_ASSOC);
        if (is_a($info, 'PEAR_Error')) {
            Horde::logMessage($info, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $info;
        }

        return String::convertCharset($info, $this->_params['charset']);
    }

    function getPriorities($type)
    {
        if (empty($type)) {
            return array();
        }
        $query = "SELECT priority_id, priority_name FROM whups_priorities WHERE type_id = $type ORDER BY priority_name";
        Horde::logMessage('SQL Query by Whups_Driver_sql::getPriorities(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);

        $priorities = $this->_db->getAssoc($query);
        if (is_a($priorities, 'PEAR_Error')) {
            Horde::logMessage($priorities, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $priorities;
        }

        return String::convertCharset($priorities, $this->_params['charset']);
    }

    function getPriorityName($type, $priority)
    {
        $query = "SELECT priority_name FROM whups_priorities WHERE type_id = $type AND priority_id = $priority";
        Horde::logMessage('SQL Query by Whups_Driver_sql::getPriorityName(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);

        $name = $this->_db->getOne($query);
        if (is_a($name, 'PEAR_Error')) {
            Horde::logMessage($name, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $name;
        }

        return String::convertCharset($name, $this->_params['charset']);
    }

    function getPriority($priorityID)
    {
        if (empty($priorityID)) {
            return false;
        }
        $query = 'SELECT priority_id, priority_name, priority_description FROM whups_priorities WHERE priority_id = ' . $this->_db->quote($priorityID);
        Horde::logMessage('SQL Query by Whups_Driver_sql::getPriority(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
        $priority = $this->_db->getAssoc($query);
        if (is_a($priority, 'PEAR_Error')) {
            Horde::logMessage($priority, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $priority;
        }

        $priority = String::convertCharset($priority, $this->_params['charset']);
        return array('id' => $priorityID,
                     'name' => isset($priority[$priorityID][0]) ? $priority[$priorityID][0] : '',
                     'description' => isset($priority[$priorityID][1]) ? $priority[$priorityID][1] : '');
    }

    function updatePriority($priorityID, $name, $description)
    {
        $query = 'UPDATE whups_priorities SET' .
                 ' priority_name = ' . $this->_db->quote(String::convertCharset($name, NLS::getCharset(), $this->_params['charset'])) . ',' .
                 ' priority_description = ' . $this->_db->quote(String::convertCharset($description, NLS::getCharset(), $this->_params['charset'])) .
                 ' WHERE priority_id = ' . $this->_db->quote($priorityID);
        Horde::logMessage('SQL Query by Whups_Driver_sql::updatePriority(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
        return $this->_db->query($query);
    }

    function deletePriority($priorityID)
    {
        $query = 'DELETE FROM whups_priorities WHERE priority_id = ' . $this->_db->quote($priorityID);
        Horde::logMessage('SQL Query by Whups_Driver_sql::deletePriority(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
        return $this->_db->query($query);
    }

    function getVersionInfo($queue)
    {
        $queue = $this->_db->quote($queue);
        $query = "SELECT version_id, version_name, version_description FROM whups_versions WHERE queue_id = $queue ORDER BY version_id";
        Horde::logMessage('SQL Query by Whups_Driver_sql::getVersionInfo(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);

        $info = $this->_db->getAll($query, null, DB_FETCHMODE_ASSOC);
        if (is_a($info, 'PEAR_Error')) {
            Horde::logMessage($info, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $info;
        }

        return String::convertCharset($info, $this->_params['charset']);
    }

    function getVersions($queue)
    {
        if (empty($queue)) {
            return array();
        }
        $queue = (int)$queue;
        $query = "SELECT version_id, version_name, version_description FROM whups_versions WHERE queue_id = $queue ORDER BY version_name DESC";
        Horde::logMessage('SQL Query by Whups_Driver_sql::getVersions(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);

        $results = $this->_db->getAssoc($query);
        if (is_a($results, 'PEAR_Error')) {
            Horde::logMessage($results, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $results;
        }

        $versions = array();
        foreach ($results as $key => $vinfo) {
            $versions[$key] = String::convertCharset($vinfo[0], $this->_params['charset']);
            if (!empty($vinfo[1])) {
                $versions[$key] .= ': ' . String::convertCharset($vinfo[1], $this->_params['charset']);
            }
        }
        return $versions;
    }

    function getVersion($versionID)
    {
        if (empty($versionID)) {
            return false;
        }
        $query = 'SELECT version_id, version_name, version_description FROM whups_versions WHERE version_id = ' . $this->_db->quote($versionID);
        Horde::logMessage('SQL Query by Whups_Driver_sql::getVersion(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
        $version = $this->_db->getAssoc($query);
        if (is_a($version, 'PEAR_Error')) {
            Horde::logMessage($version, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $version;
        }

        $version = String::convertCharset($version, $this->_params['charset']);
        return array('id' => $versionID,
                     'name' => isset($version[$versionID][0]) ? $version[$versionID][0] : '',
                     'description' => isset($version[$versionID][1]) ? $version[$versionID][1] : '');
    }

    function updateVersion($versionID, $name, $description)
    {
        $query = 'UPDATE whups_versions SET' .
                 ' version_name = ' . $this->_db->quote(String::convertCharset($name, NLS::getCharset(), $this->_params['charset'])) . ',' .
                 ' version_description = ' . $this->_db->quote(String::convertCharset($description, NLS::getCharset(), $this->_params['charset'])) .
                 ' WHERE version_id = ' . $this->_db->quote($versionID);
        Horde::logMessage('SQL Query by Whups_Driver_sql::updateVersion(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
        return $this->_db->query($query);
    }

    function deleteVersion($versionID)
    {
        $query = 'DELETE FROM whups_versions WHERE version_id = ' . $this->_db->quote($versionID);
        Horde::logMessage('SQL Query by Whups_Driver_sql::deleteVersion(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
        return $this->_db->query($query);
    }

    function getUserSearches($user)
    {
        /* We don't want to pass 'false' or an empty string to the
         * database (especially PostgreSQL, which will return a type
         * operator error), so return an empty array immediately. */
        if (empty($user)) {
            return array();
        }

        $query = "SELECT search_name, search_getvars FROM whups_users_searches WHERE user_uid = " . $this->_db->quote($user);
        Horde::logMessage('SQL Query by Whups_Driver_sql::getUserSearches(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
        $searches = $this->_db->getAssoc($query);
        if (is_a($searches, 'PEAR_Error')) {
            Horde::logMessage($searches, __FILE__, __LINE__, PEAR_LOG_ERR);
            return array();
        }

        return String::convertCharset($searches, $this->_params['charset']);
    }

    function clearUserSearches($user, $search = null)
    {
        $query = 'DELETE FROM whups_users_searches WHERE user_uid = ' . $this->_db->quote($user);
        if ($search) {
            $query .= ' AND search_name = ' . $this->_db->quote($search);
        }
        Horde::logMessage('SQL Query by Whups_Driver_sql::clearUserSearches(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
        return $this->_db->query($query);
    }

    function saveSearch($name, $getvars)
    {
        $userid  = $this->_db->quote(Auth::getAuth());
        $name    = $this->_db->quote(String::convertCharset($name, NLS::getCharset(), $this->_params['charset']));
        $getvars = $this->_db->quote($getvars);

        $this->_db->autoCommit(false);

        $query = "DELETE FROM whups_users_searches WHERE user_uid = $userid AND search_name = $name";
        Horde::logMessage('SQL Query by Whups_Driver_sql::saveSearch(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
        $r1 = $this->_db->query($query);
        $query = "INSERT INTO whups_users_searches (user_uid, search_name, search_getvars) VALUES ($userid, $name, $getvars)";
        Horde::logMessage('SQL Query by Whups_Driver_sql::saveSearch(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
        $r2 = $this->_db->query($query);
        $r3 = $this->_db->commit();

        if (is_a($r1, 'PEAR_Error')) {
            Horde::logMessage($r1, __FILE__, __LINE__, PEAR_LOG_ERR);
        }
        if (is_a($r2, 'PEAR_Error')) {
            Horde::logMessage($r2, __FILE__, __LINE__, PEAR_LOG_ERR);
        }
        if (is_a($r3, 'PEAR_Error')) {
            Horde::logMessage($r3, __FILE__, __LINE__, PEAR_LOG_ERR);
        }

        $this->_db->autoCommit(true);
    }

    function getUserQueries($user)
    {
        $query = "SELECT query_name, query_name FROM whups_users_queries WHERE user_uid = " . $this->_db->quote($user);
        Horde::logMessage('SQL Query by Whups_Driver_sql::getUserQueries(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);

        $queries = $this->_db->getAssoc($query);
        if (is_a($queries, 'PEAR_Error')) {
            Horde::logMessage($queries, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $queries;
        }

        return String::convertCharset($queries, $this->_params['charset']);
    }

    function saveQuery($name, $var)
    {
        $userid = $this->_db->quote(Auth::getAuth());
        $name   = $this->_db->quote(String::convertCharset($name, NLS::getCharset(), $this->_params['charset']));
        $var    = $this->_db->quote(String::convertCharset($var, NLS::getCharset(), $this->_params['charset']));

        $this->_db->autoCommit(false);
        $query = "DELETE FROM whups_users_queries WHERE user_uid = $userid AND query_name = $name";
        Horde::logMessage('SQL Query by Whups_Driver_sql::saveQuery(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
        $this->_db->query($query);
        $query = "INSERT INTO whups_users_queries (user_uid, query_name, query_object) VALUES ($userid, $name, $var)";
        Horde::logMessage('SQL Query by Whups_Driver_sql::saveQuery(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
        $this->_db->query($query);
        $this->_db->commit();
        $this->_db->autoCommit(true);
    }

    function loadQuery($name)
    {
        $userid = $this->_db->quote(Auth::getAuth());
        $name   = $this->_db->quote(String::convertCharset($name, NLS::getCharset(), $this->_params['charset']));

        $sql = "SELECT query_object FROM whups_users_queries WHERE user_uid = $userid AND query_name = $name";
        Horde::logMessage('SQL Query by Whups_Driver_sql::loadQuery(): ' . $sql, __FILE__, __LINE__, PEAR_LOG_DEBUG);

        $query = $this->_db->getOne($query);
        if (is_a($query, 'PEAR_Error')) {
            Horde::logMessage($query, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $query;
        }

        return String::convertCharset($query, $this->_params['charset']);
    }

    function getListeners($ticket, $withowners = true, $withrequester = true, $withresponsible = false)
    {
        $tables = 'whups_tickets_listeners l, whups_tickets t';
        $where = "(l.ticket_id = $ticket)";
        $query = "SELECT DISTINCT l.user_uid FROM $tables WHERE $where";
        Horde::logMessage('SQL Query by Whups_Driver_sql::getListeners(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
        $users = $this->_db->getCol($query);

        $tinfo = $this->getTicketDetails($ticket);
        $requester = $tinfo['user_id_requester'];
        $owners = $this->getOwners($ticket);
        if ($withresponsible) {
            $users = array_merge($users, $this->getQueueUsers($tinfo['queue']));
        }

        // Tricky - handle case where owner = requester.
        $owner_is_requester = false;
        foreach ($owners as $owner) {
            $owner = str_replace('user:', '', $owner);
            if ($owner == $requester) {
                $owner_is_requester = true;
            }
            if ($withowners) {
                $users[$owner] = $owner;
            } else {
                if (isset($users[$owner])) {
                    unset($users[$owner]);
                }
            }
        }

        if (!$withrequester) {
            if (isset($users[$requester]) && (!$withowners || $owner_is_requester)) {
                unset($users[$requester]);
            }
        } elseif (!empty($requester)) {
            $users[$requester] = $requester;
        }

        return $users;
    }

    function addAttributeDesc($type_id, $name, $desc)
    {
        // TODO: Make sure we're not adding a duplicate here (can be
        // done in the db schema).

        // FIXME: This assumes that $type is a valid type id.
        $new_id = $this->_db->nextId('whups_attributes_desc');
        if (is_a($new_id, 'PEAR_Error')) {
            Horde::logMessage($new_id, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $new_id;
        }

        $name = $this->_db->quote(String::convertCharset($name, NLS::getCharset(), $this->_params['charset']));
        $desc = $this->_db->quote(String::convertCharset($desc, NLS::getCharset(), $this->_params['charset']));

        $query = "INSERT INTO whups_attributes_desc (attribute_id, type_id, attribute_name, attribute_description) " .
                 "VALUES ($new_id, $type_id, $name, $desc)";
        Horde::logMessage('SQL Query by Whups_Driver_sql::addAttributeDesc(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
        $result = $this->_db->query($query);
        if (is_a($result, 'PEAR_Error')) {
            Horde::logMessage($result, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $result;
        }

        return $new_id;
    }

    function updateAttributeDesc($attribute_id, $newname, $newdesc)
    {
        $newname = $this->_db->quote(String::convertCharset($newname, NLS::getCharset(), $this->_params['charset']));
        $newdesc = $this->_db->quote(String::convertCharset($newdesc, NLS::getCharset(), $this->_params['charset']));

        $query = "UPDATE whups_attributes_desc SET " .
                 "attribute_name = $newname, attribute_description = $newdesc " .
                 "WHERE attribute_id = $attribute_id";
        Horde::logMessage('SQL Query by Whups_Driver_sql::updateAttributeDesc(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
        return $this->_db->query($query);
    }

    function deleteAttributeDesc($attribute_id)
    {
        // FIXME: Which one of these returns the error, or do we have
        // to check all of them for errors?
        $this->_db->autoCommit(false);
        $query = "DELETE FROM whups_attributes_desc WHERE attribute_id = $attribute_id";
        Horde::logMessage('SQL Query by Whups_Driver_sql::deleteAttributeDesc(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
        $this->_db->query($query);
        $query = "DELETE FROM whups_attributes WHERE attribute_id = $attribute_id";
        Horde::logMessage('SQL Query by Whups_Driver_sql::deleteAttributeDesc(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
        $this->_db->query($query);
        $this->_db->commit();
        $this->_db->autoCommit(true);

        return true;
    }

    function getAttributeDesc($attribute_id)
    {
        if (empty($attribute_id)) {
            return false;
        }

        $query = 'SELECT attribute_id, attribute_name, attribute_description FROM whups_attributes_desc WHERE attribute_id = ' . $this->_db->quote($attribute_id);
        Horde::logMessage('SQL Query by Whups_Driver_sql::getAttributeDesc(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
        $attribute = $this->_db->getAssoc($query);
        if (is_a($attribute, 'PEAR_Error')) {
            Horde::logMessage($attribute, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $attribute;
        }

        $attribute = String::convertCharset($attribute, $this->_params['charset']);
        return array('id' => $attribute_id,
                     'attribute_name' => isset($attribute[$attribute_id][0]) ? $attribute[$attribute_id][0] : '',
                     'attribute_description' => isset($attribute[$attribute_id][1]) ? $attribute[$attribute_id][1] : '');
    }

    function getAttributeName($attribute_id)
    {
        $attribute_id = $this->_db->quote($attribute_id);
        $query = 'SELECT attribute_name FROM whups_attributes_desc WHERE attribute_id = ' . $attribute_id;
        Horde::logMessage('SQL Query by Whups_Driver_sql::getAttributeName(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);

        $name = $this->_db->getOne($query);
        if (is_a($name, 'PEAR_Error')) {
            Horde::logMessage($name, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $name;
        }

        return String::convertCharset($name, $this->_params['charset']);
    }

    function getAttributesForType($type_id)
    {
        if (empty($type_id)) {
            return array();
        }
        $query = "SELECT attribute_id, attribute_name FROM whups_attributes_desc WHERE type_id = $type_id ORDER BY attribute_name";
        Horde::logMessage('SQL Query by Whups_Driver_sql::getAttributesForType(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);

        $attributes = $this->_db->getAssoc($query);
        if (is_a($attributes, 'PEAR_Error')) {
            Horde::logMessage($attributes, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $attributes;
        }

        return String::convertCharset($attributes, $this->_params['charset']);
    }

    function getAttributeNamesForType($type_id)
    {
        if (empty($type_id)) {
            return array();
        }
        $query = 'SELECT attribute_name FROM whups_attributes_desc WHERE type_id = ' . $type_id . ' ORDER BY attribute_name';
        Horde::logMessage('SQL Query by Whups_Driver_sql::getAttributeNamesForType(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);

        $names = $this->_db->getAll($query, null, DB_FETCHMODE_ASSOC);
        if (is_a($names, 'PEAR_Error')) {
            Horde::logMessage($names, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $names;
        }

        return String::convertCharset($names, $this->_params['charset']);
    }

    function getAttributeInfoForType($type_id)
    {
        $type_id = (int)$type_id;
        $query = "SELECT attribute_id, attribute_name, attribute_description FROM whups_attributes_desc WHERE type_id = $type_id ORDER BY attribute_id";
        Horde::logMessage('SQL Query by Whups_Driver_sql::getAttributeInfoForType(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);

        $info = $this->_db->getAll($query, null, DB_FETCHMODE_ASSOC);
        if (is_a($info, 'PEAR_Error')) {
            Horde::logMessage($info, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $info;
        }

        return String::convertCharset($info, $this->_params['charset']);
    }

    function _setAttributeValue($ticket_id, $attribute_id, $attribute_value)
    {
        $db_attribute_value = $this->_db->quote(String::convertCharset($attribute_value, NLS::getCharset(), $this->_params['charset']));

        $this->_db->autoCommit(false);
        $query = "DELETE FROM whups_attributes WHERE ticket_id = $ticket_id AND attribute_id = $attribute_id";
        Horde::logMessage('SQL Query by Whups_Driver_sql::_setAttributeValue(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
        $this->_db->query($query);

        if (!empty($attribute_value)) {
            $query = "INSERT INTO whups_attributes " .
                     "(ticket_id, attribute_id, attribute_value) VALUES " .
                     "($ticket_id, $attribute_id, $db_attribute_value)";
            Horde::logMessage('SQL Query by Whups_Driver_sql::_setAttributeValue(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
            $this->_db->query($query);
        }

        $this->_db->commit();
        $this->_db->autoCommit(true);
    }

    function getTicketAttributes($ticket_id)
    {
        $query = 'SELECT attribute_id, attribute_value FROM whups_attributes WHERE ticket_id = ' . $ticket_id;
        Horde::logMessage('SQL Query by Whups_Driver_sql::getTicketAttributes(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);

        $attributes = $this->_db->getAssoc($query);
        if (is_a($attributes, 'PEAR_Error')) {
            Horde::logMessage($attributes, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $attributes;
        }

        return String::convertCharset($attributes, $this->_params['charset']);
    }

    function getTicketAttributesWithNames($ticket_id)
    {
        $ticket_id = (int)$ticket_id;
        $query = 'SELECT d.attribute_name, a.attribute_value ' .
                 'FROM whups_attributes a INNER JOIN whups_attributes_desc d ON (d.attribute_id = a.attribute_id)' .
                 "WHERE a.ticket_id = $ticket_id " .
                 'ORDER BY d.attribute_name';
        Horde::logMessage('SQL Query by Whups_Driver_sql::getTicketAttributesWithNames(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);

        $attributes = $this->_db->getAssoc($query);
        if (is_a($attributes, 'PEAR_Error')) {
            Horde::logMessage($attributes, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $attributes;
        }

        return String::convertCharset($attributes, $this->_params['charset']);
    }

    function getAllTicketAttributesWithNames($ticket_id)
    {
        $query = 'SELECT d.attribute_name, a.attribute_value ' .
                 'FROM whups_attributes_desc d ' .
                 "LEFT JOIN whups_tickets t ON (t.ticket_id = $ticket_id) " .
                 "LEFT OUTER JOIN whups_attributes a ON (d.attribute_id = a.attribute_id AND a.ticket_id = $ticket_id) " .
                 'WHERE d.type_id = t.type_id ' .
                 'ORDER BY d.attribute_name';
        Horde::logMessage('SQL Query by Whups_Driver_sql::getAllTicketAttributesWithNames(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);

        $attributes = $this->_db->getAssoc($query);
        if (is_a($attributes, 'PEAR_Error')) {
            Horde::logMessage($attributes, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $attributes;
        }

        return String::convertCharset($attributes, $this->_params['charset']);
    }

    function getOwners($ticketId)
    {
        if (is_array($ticketId)) {
            // No need to run a query for an empty array, and it would
            // result in an invalid SQL query anyway.
            if (!count($ticketId)) {
                return array();
            }

            $query = 'SELECT ticket_id AS id, ticket_owner AS owner FROM whups_ticket_owners WHERE ticket_id IN (' . implode(',', $ticketId) . ')';
            Horde::logMessage('SQL Query by Whups_Driver_sql::getOwners(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
            return $this->_db->getAll($query, DB_FETCHMODE_ASSOC);
        } else {
            $query = 'SELECT ticket_owner, ticket_owner FROM whups_ticket_owners WHERE ticket_id = ' . (int)$ticketId;
            Horde::logMessage('SQL Query by Whups_Driver_sql::getOwners(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
            return $this->_db->getAssoc($query);
        }
    }

    function updateLog($ticket_id, $user, $changes = array(), $transactionId = null)
    {
        if (is_null($transactionId)) {
            $transactionId = $this->newTransaction();
            if (is_a($transactionId, 'PEAR_Error')) {
                return $transactionId;
            }
        }

        foreach ($changes as $type => $value) {
            $log_id = $this->_db->nextId('whups_logs');
            if (is_a($log_id, 'PEAR_Error')) {
                return $log_id;
            }

            $query = 'INSERT INTO whups_logs (log_id, transaction_id, ticket_id, log_timestamp, user_id, log_type, log_value, log_value_num) ' .
                sprintf('VALUES (%d, %d, %d, %d, %s, %s, %s, %d)',
                        (int)$log_id,
                        (int)$transactionId,
                        (int)$ticket_id,
                        time(),
                        $this->_db->quote((string)$user),
                        $this->_db->quote($type),
                        $this->_db->quote(String::convertCharset($value, NLS::getCharset(), $this->_params['charset'])),
			(int)$value);
            Horde::logMessage('SQL Query by Whups_Driver_sql::updateLog(): ' . $query, __FILE__, __LINE__, PEAR_LOG_DEBUG);
            $result = $this->_db->query($query);
            if (is_a($result, 'PEAR_Error')) {
                Horde::logMessage($result, __FILE__, __LINE__, PEAR_LOG_ERR);
                return $result;
            }
        }

        return $transactionId;
    }

    function newTransaction()
    {
        return $this->_db->nextId('whups_transactions');
    }

    /**
     * @access private
     */
    function initialise()
    {
        Horde::assertDriverConfig($this->_params, 'tickets',
            array('phptype', 'hostspec', 'username', 'database'));

        /* Connect to the SQL server using the supplied parameters. */
        require_once 'DB.php';
        $this->_db = &DB::connect($this->_params,
                                  array('persistent' => !empty($this->_params['persistent'])));
        if (is_a($this->_db, 'PEAR_Error')) {
            Horde::fatal($this->_db, __FILE__, __LINE__);
        }

        $this->_db->setOption('portability', DB_PORTABILITY_LOWERCASE | DB_PORTABILITY_ERRORS);

        return true;
    }

    function _generateWhere($table, $fields, &$info, $type)
    {
        $where = '';
        $this->_mapFields($info);

        foreach ($fields as $field) {
            if (isset($info[$field])) {
                $prop = $info[$field];
                if (is_array($info[$field])) {
                    $clauses = array();
                    foreach ($prop as $pprop) {
                        if (@settype($pprop, $type)) {
                            $clauses[] = "$table.$field = " . $this->_db->quote($pprop);
                        }
                    }
                    if (count($clauses)) {
                        $where = $this->_addWhere($where, true, implode(' OR ', $clauses));
                    }
                } else {
                    $success = @settype($prop, $type);
                    $where = $this->_addWhere($where, !is_null($prop) && $success, "$table.$field = " . $this->_db->quote($prop));
                }
            }
        }

        foreach ($fields as $field) {
            if (isset($info["not$field"])) {
                $prop = $info["not$field"];

                if (strpos($prop, ',') === false) {
                    $success = @settype($prop, $type);
                    $where = $this->_addWhere($where, $prop && $success, "$table.$field <> " . $this->_db->quote($prop));
                } else {
                    $set = explode(',', $prop);

                    foreach ($set as $prop) {
                        $success = @settype($prop, $type);
                        $where = $this->_addWhere($where, $prop && $success, "$table.$field <> " . $this->_db->quote($prop));
                    }
                }
            }
        }

        return $where;
    }

    function _mapFields(&$info)
    {
        foreach ($info as $key => $val) {
            if ($key === 'id') {
                $info['ticket_id'] = $info['id'];
                unset($info['id']);
            } elseif ($key === 'state' ||
                      $key === 'type' ||
                      $key === 'queue' ||
                      $key === 'priority') {
                $info[$key . '_id'] = $info[$key];
                unset($info[$key]);
            } elseif ($key === 'requester') {
                $info['user_id_' . $key] = $info[$key];
                unset($info[$key]);
            }
        }
    }

    function _addWhere($where, $condition, $clause, $conjunction = 'AND')
    {
        if (!empty($condition)) {
            if (!empty($where)) {
                $where .= " $conjunction ";
            }

            $where .= "($clause)";
        }

        return $where;
    }

    function _prefixTableToColumns($table, $columns)
    {
        $join = "";

        $clause = '';
        foreach ($columns as $column) {
            $clause .= "$join$table.$column";
            $join = ', ';
        }

        return $clause;
    }

}
