<?php
/**
 * $Horde: whups/lib/Query.php,v 1.34 2005/01/03 14:35:44 jan Exp $
 *
 * Copyright 2001-2005 Robert E. Coyle <robertecoyle@hotmail.com>
 *
 * See the enclosed file LICENSE for license information (BSD). If you
 * did not receive this file, see http://www.horde.org/licenses/bsdl.php.
 */

/**
 * Mode types.
 */
define('QUERY_TYPE_AND',       1);
define('QUERY_TYPE_OR',        2);
define('QUERY_TYPE_NOT',       3);
define('QUERY_TYPE_CRITERION', 4);

/**
 * Criterion types.
 */
define('CRITERION_ID',             1);
define('CRITERION_QUEUE',          2);
define('CRITERION_TYPE',           3);
define('CRITERION_STATE',          4);
define('CRITERION_PRIORITY',       5);
define('CRITERION_OWNERS',         7);
define('CRITERION_GROUPS',         9);
define('CRITERION_REQUESTER',      8);
define('CRITERION_ADDED_COMMENT', 11);
define('CRITERION_COMMENT',       12);
define('CRITERION_SUMMARY',       13);
define('CRITERION_ATTRIBUTE',     14);

/**
 * Operators for integer fields.
 */
define('OPERATOR_GREATER', 1);
define('OPERATOR_LESS',    2);
define('OPERATOR_EQUAL',   3);

/**
 * Operators for text fields.
 */
define('OPERATOR_CI_SUBSTRING',  4);
define('OPERATOR_CS_SUBSTRING',  5);
define('OPERATOR_WORD',          6);
define('OPERATOR_PATTERN',       7);

$textOperators = array(
    OPERATOR_EQUAL        => _("Exact Match"),
    OPERATOR_CI_SUBSTRING => _("Case Insensitive Substring"),
    // OPERATOR_CS_SUBSTRING => _("Case Sensitive Substring"),
    OPERATOR_WORD         => _("Match Word"),
    OPERATOR_PATTERN      => _("Match Pattern"));

/**
 * array(
 *     'type'      => QUERY_TYPE_...
 *     'children'  => array(...) unless type == QUERY_TYPE_CRITERION
 *     'criterion' => CRITERION_... if  type == QUERY_TYPE_CRITERION
 *     'operator'  => OPERATOR_...  if  type == QUERY_TYPE_CRITERION
 *     'value'     => other argument to operator of criterion
 *     'ttype'     => associated ticket type of criterion
 */

/**
 * @package Whups
 */
class Query {

    var $_currentTicketType;
    var $_parameters;
    var $_query;

    function Query()
    {
        global $whups;

        $types = $whups->getAllTypes();
        $ttype = array_shift(array_keys($types));

        $this->setCurrentTicketType($ttype);
        $this->reset();
    }

    /**
     * Reset the query to the default empty query.
     */
    function reset()
    {
        $and = array(
            'type'     => QUERY_TYPE_AND,
            'children' => array());

        $this->_query = $and;
        $this->_parameters = array();
    }

    function getCurrentTicketType()
    {
        return $this->_currentTicketType;
    }

    function setCurrentTicketType($type)
    {
        $this->_currentTicketType = $type;
    }

    /**
     * Parameter form.
     */
    function getParameterForm(&$vars)
    {
        if (count($this->_parameters) == 0)
            return null;

        $result = new Horde_Form($vars, _("Query Parameters"), 'queryparameters');

        // You can resubmit this as many times as you like.
        $result->_useFormToken = false;

        foreach ($this->_parameters as $index => $name) {
            $result->addVariable($name, $name, 'text', true);
        }

        return $result;
    }

    /**
     * Operations.
     */
    function delete($pathstring)
    {
        $path = Query::stringToPath($pathstring);
        $qobj = &$this->_query;

        if (strlen($pathstring) == 0) {
            // Deleting the root node is equivalent to creating a new
            // blank query.
            $this->reset();
        } else {
            $count = count($path) - 1;
            for ($i = 0; $i < $count; $i++) {
                $qobj = &$qobj['children'][$path[$i]];
            }

            if (!empty($qobj['children'][$path[$count]]['value']) &&
                $this->_getParameterName($qobj['children'][$path[$count]]['value']) !== null) {
                unset($this->_parameters[array_search($pn, $this->_parameters)]);
            }

            array_splice($qobj['children'], $path[$count], 1);
        }
    }

    function hoist($pathstring)
    {
        $path = Query::stringToPath($pathstring);
        $qobj = &$this->_query;

        if (strlen($pathstring) == 0) {
            // Can't hoist the root node.
        } else {
            $count = count($path) - 1;

            for ($i = 0; $i < $count; $i++) {
                $qobj = &$qobj['children'][$path[$i]];
            }

            $cobj = &$qobj['children'][$path[$count]];

            // TODO: make sure we're hoisting a branch.
            array_splice($qobj['children'], $path[$count], 0, $cobj['children']);
            array_splice($cobj['children'], 0, count($cobj['children']));
        }
    }

    function insertBranch($pathstring, $type)
    {
        $path = Query::stringToPath($pathstring);
        $qobj = &$this->_query;

        $newbranch = array(
            'type'     => $type,
            'children' => array());

        $count = count($path);

        for ($i = 0; $i < $count; $i++) {
            $qobj = &$qobj['children'][$path[$i]];
        }

        if (!isset($qobj['children']) ||
            !is_array($qobj['children'])) {
            $qobj['children'] = array();
        }

        array_push($qobj['children'], $newbranch);
        array_push($path, count($qobj['children']) - 1);

        return Query::pathToString($path);
    }

    function insertCriterion($pathstring, $criterion, $cvalue, $operator, $value)
    {
        $path = Query::stringToPath($pathstring);
        $qobj = &$this->_query;

        $value = trim($value);
        if ($value{0} == '"') {
            // FIXME: The last character should be '"' as well.
            $value = substr($value, 1, -1);
        } else {
            $pn = $this->_getParameterName($value);
            if ($pn !== null)
                $this->_parameters[] = $pn;
        }

        $newbranch = array(
            'type'      => QUERY_TYPE_CRITERION,
            'criterion' => $criterion,
            'cvalue'    => $cvalue,
            'operator'  => $operator,
            'value'     => $value,
            'ttype'     => $this->_currentTicketType);

        $count = count($path);
        for ($i = 0; $i < $count; $i++) {
            $qobj = &$qobj['children'][$path[$i]];
        }

        array_push($qobj['children'], $newbranch);
    }

    /**
     * Top down traversal.
     */
    function walk(&$obj, $method)
    {
        $path = array();
        $more = array();
        $this->_walk($this->_query, $more, $path, $obj, $method);
    }

    function _walk(&$node, &$more, &$path, &$obj, $method)
    {
        if ($node['type'] == QUERY_TYPE_CRITERION) {
            $obj->$method($more, $path, QUERY_TYPE_CRITERION, $node['ttype'], $node['criterion'],
                          $node['cvalue'], $node['operator'], $node['value']);
        } else {
            $obj->$method($more, $path, $node['type'], null, null, null, null, null);
        }

        if (isset($node['children'])) {
            $count = count($node['children']);

            for ($i = 0; $i < $count; $i++) {
                array_push($path, $i);
                array_push($more, ($i < $count - 1));
                $this->_walk($node['children'][$i], $more, $path, $obj, $method);
                array_pop($more);
                array_pop($path);
            }
        }
    }

    /**
     * Bottom up traversal.
     */
    function reduce(&$obj, $method, &$vars)
    {
        return $this->_reduce($this->_query, $obj, $method, $vars);
    }

    function _reduce(&$node, &$obj, $method, &$vars)
    {
        $args = array();

        if (isset($node['children'])) {
            $count = count($node['children']);

            for ($i = 0; $i < $count; $i++) {
                $result = $this->_reduce($node['children'][$i], $obj, $method, $vars);
                array_push($args, $result);
            }
        }

        if ($node['type'] == QUERY_TYPE_CRITERION) {
            $value = $node['value'];

            $pn = $this->_getParameterName($value);
            if ($pn !== null) {
                $value = $vars->get($pn);
            }

            $result = $obj->$method($args, QUERY_TYPE_CRITERION, $node['ttype'], $node['criterion'],
                                    $node['cvalue'], $node['operator'], $value);
        } else {
            $result = $obj->$method($args, $node['type'], null, null, null, null, null);
        }

        return $result;
    }

    /**
     * Saving / retrieving queries.
     */
    function save($name)
    {
        global $whups;

        $obj = serialize(array($this->_currentTicketType, $this->_query, $this->_parameters));

        $whups->saveQuery($name, $obj);
    }

    function load($name)
    {
        global $whups;

        $this->reset();

        $obj = $whups->loadQuery($name);
        $obj = unserialize($obj);

        $this->_currentTicketType = $obj[0];
        $this->_query = $obj[1];
        $this->_parameters = $obj[2];

        return true;
    }

    /**
     * @access private
     */
    function _getParameterName($value)
    {
        if (strcmp(substr($value, 0, 2), '${'))
            return null;

        $pn = substr($value, 2, -1);
        if (!is_string($pn))
            $pn = null;

        return $pn;
    }

    /**
     * @static
     */
    function pathToString(&$path)
    {
        return implode(',', $path);
    }

    /**
     * @static
     */
    function stringToPath($pathstring)
    {
        if (strlen($pathstring) == 0) {
            return array();
        }

        return explode(',', $pathstring);
    }

}

/**
 * Forms for editing queries.
 *
 * @package Whups
 */
class InsertBranchForm extends Horde_Form {

    function InsertBranchForm(&$vars)
    {
        parent::Horde_Form($vars, _("Insert Branch"));

        $branchtypes = array(
            QUERY_TYPE_AND => _("And"),
            QUERY_TYPE_OR  => _("Or"),
            QUERY_TYPE_NOT => _("Not"));

        $this->addHidden(_("Path"), 'path', 'text', false, true);
        $this->addVariable(_("Branch Type"), 'type', 'enum', true, false, null, array($branchtypes));
    }

}

/**
 * @package Whups
 */
class ChangeTypeForm extends Horde_Form {

    function ChangeTypeForm(&$vars)
    {
        global $whups;

        parent::Horde_Form($vars, _("Change Ticket Type"));

        $this->addVariable(_("Current Type"), 'ttype', 'enum', true, false, null, array($whups->getAllTypes()));
    }

    /**
     * We want to render in a slightly modified form.
     */
    function &getRenderer()
    {
        return $r = &new QuerySetCurrentTypeRenderer();
    }

    function execute(&$vars)
    {
        global $whups_query;
        $whups_query->setCurrentTicketType($vars->get('ttype'));
    }

}

/**
 * @package Whups
 */
class UserCriterionForm extends Horde_Form {

    function UserCriterionForm(&$vars)
    {
        global $textOperators;
        global $whups;

        parent::Horde_Form($vars, _("Add User Criterion"));

        $this->addVariable(_("User ID"), 'user', 'text', true);
        $this->addVariable(_("Match Operator"), 'operator', 'enum', true, false, null, array($textOperators));
        $this->addVariable(_("Search Owners"), 'owners', 'boolean', false);
        $this->addVariable(_("Search Requester"), 'requester', 'boolean', false);
        $this->addVariable(_("Search Comments"), 'comments', 'boolean', false);
    }

    function execute(&$vars)
    {
        global $whups_query;
        global $whups;

        $path = $vars->get('path');
        $user = $vars->get('user');
        $operator = $vars->get('operator');
        $owners = $vars->get('owners');
        $requester = $vars->get('requester');
        $comments = $vars->get('comments');

        // If we're adding more than one criterion, put them all under
        // an OR node (which should be what is wanted in the general
        // case).
        if ((int)$owners + (int)$requester + (int)$comments > 1) {
            $path = $whups_query->insertBranch($path, QUERY_TYPE_OR);
        }

        if ($owners) {
            $whups_query->insertCriterion($path, CRITERION_OWNERS, null, $operator, $user);
        }

        if ($requester) {
            $whups_query->insertCriterion($path, CRITERION_REQUESTER, null, $operator, $user);
        }

        if ($comments) {
            $whups_query->insertCriterion($path, CRITERION_ADDED_COMMENT, null, $operator, $user);
        }

        $this->unsetVars($vars);
    }

}

/**
 * @package Whups
 */
class GroupCriterionForm extends Horde_Form {

    function GroupCriterionForm(&$vars)
    {
        global $textOperators;
        global $whups;

        parent::Horde_Form($vars, _("Add Group Criterion"));

        $groups = &Group::singleton();
        $grouplist = $groups->listGroups();
        if (is_a($grouplist, 'PEAR_Error') || !count($grouplist)) {
            $type_params = array(_("Could not find any groups."));
            $this->addVariable(_("Groups"), 'groups', 'invalid', false, false, null, $type_params);
        } else {
            $this->addVariable(_("Groups"), 'groups', 'enum', true, false, null, array($grouplist));
        }
    }

    function execute(&$vars)
    {
        global $whups_query;
        global $whups;

        $path = $vars->get('path');
        $groups = $vars->get('groups');

        if ($groups) {
            $whups_query->insertCriterion($path, CRITERION_GROUPS, null, OPERATOR_EQUAL, $groups);
        }

        $this->unsetVars($vars);
    }

}

/**
 * @package Whups
 */
class TextCriterionForm extends Horde_Form {

    function TextCriterionForm(&$vars)
    {
        global $textOperators;

        parent::Horde_Form($vars, _("Add Text Criterion"));

        $this->addVariable(_("Text"), 'text', 'text', true);
        $this->addVariable(_("Match Operator"), 'operator', 'enum', true, false, null, array($textOperators));
        $this->addVariable(_("Search Summary"), 'summary', 'boolean', false);
        $this->addVariable(_("Search Comments"), 'comments', 'boolean', false);
    }

    function execute(&$vars)
    {
        global $whups_query;

        $path = $vars->get('path');
        $text = $vars->get('text');
        $operator = $vars->get('operator');
        $summary = $vars->get('summary');
        $comments = $vars->get('comments');

        if ($summary && $comments) {
            $path = $whups_query->insertBranch($path, QUERY_TYPE_OR);
        }

        if ($summary) {
            $whups_query->insertCriterion($path, CRITERION_SUMMARY, null, $operator, $text);
        }

        if ($comments) {
            $whups_query->insertCriterion($path, CRITERION_COMMENT, null, $operator, $text);
        }

        $this->unsetVars($vars);
    }

}

/**
 * @package Whups
 */
class PropertyCriterionForm extends Horde_Form {

    function PropertyCriterionForm(&$vars)
    {
        global $whups, $whups_query;

        parent::Horde_Form($vars, _("Add Property Criterion"));

        $this->addVariable(_("Id"), 'id', 'intlist', false);

        /* Queues. */
        $queues = Whups::permissionsFilter($whups->getQueues(), 'queue', PERMS_READ);
        if (count($queues)) {
            $this->addVariable(_("Queue"), 'queue', 'enum', false, false, null, array($queues, _("Any")));
        }

        /* Types - this is just a toggle to add a hard filter on the
         * current ticket type. */
        $type = $whups_query->getCurrentTicketType();
        $this->addVariable(_("Type"), 'ttype', 'enum', false, false,
                           _("Use the Change Ticket Type form to switch the current Ticket Type. If you select \"Any\" here, the criteria you add will search every kind of ticket. If you choose the current ticket type, you will be limiting this branch of your search to that type."),
                           array(array($type => $whups->getTypeName($type)), _("Any")));

        /* States. */
        $type = $whups_query->getCurrentTicketType();
        $this->addVariable(_("State"), 'state', 'enum', false, false, null, array($whups->getStates($type), _("Any")));

        /* Priorities. */
        $this->addVariable(_("Priority"), 'priority', 'enum', false, false, null, array($whups->getPriorities($type), _("Any")));
    }

    function execute(&$vars)
    {
        global $whups_query;

        $path = $vars->get('path');

        $id = $vars->get('id');
        if (trim($id) != '') {
            $newpath = $path;
            $ids = split("[\\t\\n ,]+", $id);

            if (count($ids) > 1) {
                $newpath = $whups_query->insertBranch($path, QUERY_TYPE_OR);
            }

            foreach ($ids as $id) {
                $whups_query->insertCriterion($newpath, CRITERION_ID, null, OPERATOR_EQUAL, $id);
            }
        }

        $queue = $vars->get('queue');
        if ($queue) {
            $whups_query->insertCriterion($path, CRITERION_QUEUE, null, OPERATOR_EQUAL, $queue);
        }

        $type = $vars->get('ttype');
        if ($type) {
            $whups_query->insertCriterion($path, CRITERION_TYPE, null, OPERATOR_EQUAL, $type);
        }

        $state = $vars->get('state');
        if ($state) {
            $whups_query->insertCriterion($path, CRITERION_STATE, null, OPERATOR_EQUAL, $state);
        }

        $priority = $vars->get('priority');
        if ($priority) {
            $whups_query->insertCriterion($path, CRITERION_PRIORITY, null, OPERATOR_EQUAL, $priority);
        }

        $this->unsetVars($vars);
    }

}

/**
 * @package Whups
 */
class AttributeCriterionForm extends Horde_Form {

    function AttributeCriterionForm(&$vars)
    {
        global $textOperators, $whups, $whups_query;

        parent::Horde_Form($vars, _("Add Attribute Criterion"));

        $type = $whups_query->getCurrentTicketType();
        $this->attribs = $whups->getAttributesForType($type);

        if (count($this->attribs) > 0) {
            $this->addVariable(_("Match"), 'text', 'text', true);
            $this->addVariable(_("Match Operator"), 'operator', 'enum', true, false, null, array($textOperators));

            foreach ($this->attribs as $id => $name) {
                $this->addVariable(sprintf(_("Search %s Attribute"), $name), "a$id", 'boolean', false);
            }
        } else {
            $this->addVariable(_("Search Attribute"), 'attribute', 'invalid', true, false, null, array(_("There are no attributes defined for the current type")));
        }
    }

    function execute(&$vars)
    {
        global $whups_query;
        global $whups;

        $path = $vars->get('path');
        $text = $vars->get('text');
        $operator = $vars->get('operator');

        $count = 0;

        foreach ($this->attribs as $id => $name) {
            $count += $vars->exists("a$id") ? 1 : 0;
        }

        if ($count > 1) {
            $path = $whups_query->insertBranch($path, QUERY_TYPE_OR);
        }

        foreach ($this->attribs as $id => $name) {
            if ($vars->exists("a$id")) {
                $whups_query->insertCriterion($path, CRITERION_ATTRIBUTE, $id, $operator, $text);
            }
        }

        $this->unsetVars($vars);
    }

}

/**
 * @package Whups
 */
class ChooseQueryNameForSaveForm extends Horde_Form {

    function ChooseQueryNameForSaveForm(&$vars)
    {
        global $whups;

        parent::Horde_Form($vars, _("Save Query"));

        $this->addVariable(_("Name"), 'name', 'text', true);
    }

    function execute(&$vars)
    {
        global $whups_query;
        global $whups;

        $whups_query->save($vars->get('name'));

        $this->unsetVars($vars);
    }

}

/**
 * @package Whups
 */
class ChooseQueryNameForLoadForm extends Horde_Form {

    function ChooseQueryNameForLoadForm(&$vars)
    {
        global $whups;

        parent::Horde_Form($vars, _("Load Query"));

        $query_array = $whups->getUserQueries(Auth::getAuth());
        if (count($query_array)) {
            $queries = 'enum';
            $type_params = array($query_array);
        } else {
            $queries = 'invalid';
            $type_params = array(_("You have no saved queries."));
        }

        $this->addVariable(_("Name"), 'name', $queries, true, false, null, $type_params);
    }

    function execute(&$vars)
    {
        global $whups_query;
        global $whups;

        $whups_query->load($vars->get('name'));

        $this->unsetVars($vars);
    }

}
