<?php
/**
 * Whups external API interface.
 *
 * $Horde: whups/lib/api.php,v 1.76 2004/11/04 18:14:39 chuck Exp $
 *
 * This file defines Whups' external API interface. Other applications
 * can interact with Whups through this API.
 *
 * @package Whups
 */

$_services['perms'] = array(
    'args' => array(),
    'type' => 'stringArray'
);

$_services['browse'] = array(
    'args' => array('path'),
    'type' => 'stringArray',
);

$_services['list'] = array(
    'args' => array(),
    'type' => 'stringArray'
);

$_services['add'] = array(
    'args' => array('name'),
    'type' => 'integer'
);

$_services['summary'] = array(
    'args' => array(),
    'type' => 'string'
);

$_services['get_assigned_ticket_ids'] = array(
    'args' => array(),
    'type' => 'stringArray'
);

$_services['get_requested_ticket_ids'] = array(
    'args' => array(),
    'type' => 'stringArray'
);

$_services['get_schema'] = array(
    'args' => array(),
    'type' => 'stringArray'
);

$_services['add_tickets'] = array(
    'args' => array('info'), // info is a keyed array of lots of stuff - is this meant to reflect that?
    'type' => 'stringArray'
);

$_services['set_ticket_attributes'] = array(
    'args' => array('info'), // info is a keyed array of lots of stuff - is this meant to reflect that?
    'type' => 'stringArray'
);

$_services['query'] = array(
    'args' => array('info'),
    'type' => 'stringArray'
);

$_services['admin_list'] = array(true);

$_services['getListTypes'] = array(
    'args' => array(),
    'type' => 'stringArray'
);

$_services['listAs'] = array(
    'args' => array('type'),
    'type' => 'stringArray'
);

$_services['getQueueDetails'] = array(
    'args' => array('queue_id'),
    'type' => 'stringArray'
);

$_services['getTicketDetails'] = array(
    'args' => array('queue_id', 'state'),
    'type' => 'stringArray'
);


function _whups_admin_list()
{
    return array('admin' =>
                 array('link' => '%application%/admin.php',
                       'icon' => 'admin.png'));
}

/**
 * Browse through Whups' object tree.
 *
 * @param string $path  The level of the tree to browse.
 *
 * @return array  The contents of $path
 */
function _whups_browse($path = '')
{
    require_once dirname(__FILE__) . '/base.php';
    global $whups, $registry;

    if (substr($path, 0, 5) == 'whups') {
        $path = substr($path, 5);
    }
    if (substr($path, 0, 1) == '/') {
        $path = substr($path, 1);
    }
    if (substr($path, -1) == '/') {
        $path = substr($path, 0, -1);
    }

    if (empty($path)) {
        $results = array(
            'whups/queue' => array('name' => _("Queues"),
                             'icon' => $registry->getImageDir() . '/whups.png',
                             'browseable' => count(Whups::permissionsFilter($whups->getQueues(), 'queue', PERMS_READ))),
        );
    } else {
        $path = explode('/', $path);
        $queues = Whups::permissionsFilter($whups->getQueues(), 'queue', PERMS_READ);
        $results = array();

        switch ($path[0]) {
        case 'queue':
            if (count($path) == 1) {
                foreach ($queues as $queue => $name) {
                    $results['whups/queue/' . $queue] = array('name' => $name,
                                                              'browseable' => true);
                }
            } else {
                if (!isset($queues[$path[1]])) {
                    return PEAR::raiseError('permission denied');
                }

                $tickets = $whups->getTicketsByProperties(array('queue' => $path[1]));
                foreach ($tickets as $ticket) {
                    $results['whups/ticket/' . $ticket['id']] = array('name' => $ticket['summary'],
                                                                      'browseable' => false);
                }
            }
            break;
        }
    }

    return $results;
}

function _whups_list()
{
    require_once dirname(__FILE__) . '/base.php';
    return Whups::permissionsFilter($GLOBALS['whups']->getQueues(), 'queue', PERMS_READ);
}

function _whups_add($name)
{
    require_once dirname(__FILE__) . '/base.php';
    if (Auth::isAdmin('whups:admin')) {
        return $GLOBALS['whups']->addQueue($name, '');
    } else {
        return PEAR::raiseError('You must be an administrator to perform this action.');
    }
}

function _whups_get_assigned_ticket_ids()
{
    require_once dirname(__FILE__) . '/base.php';
    global $whups;

    $info = array('owner' => 'user:' . Auth::getAuth(),
                  'nores' => true);
    $tickets = &$whups->getTicketsByProperties($info);
    if (is_a($tickets, 'PEAR_Error')) {
        return $tickets;
    }
    $result = array();
    foreach ($tickets as $ticket) {
        $result[] = $ticket['id'];
    }
    return $result;
}

function _whups_get_requested_ticket_ids()
{
    require_once dirname(__FILE__) . '/base.php';
    global $whups;

    $info = array('requester' => Auth::getAuth(),
                  'nores'     => true);
    $tickets = &$whups->getTicketsByProperties($info);
    if (is_a($tickets, 'PEAR_Error')) {
        return $tickets;
    }
    $result = array();
    foreach ($tickets as $ticket) {
        $result[] = (int) $ticket['id'];
    }
    return $result;
}

function _whups_get_schema()
{
    require_once dirname(__FILE__) . '/base.php';
    global $whups;

    /// First, load all the type info
    $types = $whups->getAllTypeInfo();
    if (is_a($types, 'PEAR_Error')) {
        return $types;
    }

    $c = count($types);
    for ($i = 0; $i < $c; $i++) {
        // Hmm.. the ids are strings from the database - patch them up
        // here.
        settype($types[$i]['type_id'], 'int');

        // Now load in the state info.
        $states = $whups->getAllStateInfo($types[$i]['type_id']);
        if (is_a($states, 'PEAR_Error')) {
            return $states;
        }

        $ci = count($states);
        for ($j = 0; $j < $ci; $j++) {
            settype($states[$j]['state_id'], 'int');
        }

        // Now get the priorities.
        $priorities = $whups->getAllPriorityInfo($types[$i]['type_id']);
        if (is_a($priorities, 'PEAR_Error')) {
            return $priorities;
        }

        $ci = count($priorities);
        for ($j = 0; $j < $ci; $j++) {
            settype($priorities[$j]['priority_id'], 'int');
        }

        // And the attribute descriptors
        $attrib_descs = $whups->getAttributeInfoForType($types[$i]['type_id']);
        if (is_a($attrib_descs, 'PEAR_Error')) {
            return $attrib_descs;
        }

        $ci = count($attrib_descs);
        for ($j = 0; $j < $ci; $j++) {
            settype($attrib_descs[$j]['attribute_id'], 'int');
        }

        $types[$i]['type_states']     = $states;
        $types[$i]['type_priorities'] = $priorities;
        $types[$i]['type_attributes'] = $attrib_descs;
    }

    // Now load all the queue info.
    $queues = $whups->getAllQueueInfo();
    if (is_a($queues, 'PEAR_Error')) {
        return $queues;
    }

    $c = count($queues);
    for ($i = 0; $i < $c; $i++) {
        settype($queues[$i]['queue_id'], 'int');
        settype($queues[$i]['queue_subjectlist'], 'int');
        settype($queues[$i]['queue_versioned'], 'int');

        $mid = $queues[$i]['queue_id'];

        $type_ids_r = $whups->getTypeIds($mid);
        $type_ids   = array();
        $ci = count($type_ids_r);
        for ($j = 0; $j < $ci; $j++)
            $type_ids[$j] = (int) $type_ids_r[$j]['type_id'];

        $versions = $whups->getVersionInfo($mid);
        $ci = count($versions);
        for ($j = 0; $j < $ci; $j++) {
            settype($versions[$j]['version_id'], 'int');
        }

        $queues[$i]['queue_type_ids'] = $type_ids;
        $queues[$i]['queue_versions'] = $versions;
    }

    // And return the results.
    return array('types' => $types, 'queues' => $queues);
}

function _whups_add_tickets($ticket_info_array)
{
    require_once dirname(__FILE__) . '/base.php';
    require_once dirname(__FILE__) . '/Create.php';
    global $whups;

    if (!is_array($ticket_info_array))
        return PEAR::raiseError('Invalid arguments');

    $attribcache = array();

    // Validate the parameters we've been given.
    $tinfo  = array();
    $tainfo = array();

    foreach ($ticket_info_array as $index => $cinfo) {
        $vars = &new Variables($cinfo);

        $form1 = &new CreateStep1Form($vars);
        $form2 = &new CreateStep2Form($vars);
        $form3 = &new CreateStep3Form($vars);
        $form4 = &new CreateStep4Form($vars);

        // FIXME: This is an almighty hack, but we can't have form
        // tokens in rpc calls.
        $form1->_useFormToken = false;
        $form2->_useFormToken = false;
        $form3->_useFormToken = false;
        $form4->_useFormToken = false;

        $form1->validate($vars);
        $form2->validate($vars);
        $form3->validate($vars);

        // Complain if we've been given bad parameters.
        if (!$form1->isValid() || !$form2->isValid() || !$form3->isValid()) {
            $f1 = implode(', ', $form1->_errors);
            $f2 = implode(', ', $form2->_errors);
            $f3 = implode(', ', $form3->_errors);
            $v1 = implode(', ', array_keys($vars->_vars[0]));
            $v2 = implode(', ', array_values($vars->_vars[0]));

            return PEAR::raiseError("Invalid arguments 2 $index ($f1) ($f2) ($f3) ($v1) ($v2)");
        }

        $tinfo[$index] = array();
        $form1->getInfo($vars, $tinfo[$index]);
        $form2->getInfo($vars, $tinfo[$index]);
        $form3->getInfo($vars, $tinfo[$index]);

        // More checks if we're assigning the ticket at create-time.
        if (Auth::getAuth() && $whups->isCategory('assigned', $vars->get('state'))) {
            $form4->validate($vars);

            if (!$form4->isValid()) {
                return PEAR::raiseError('Invalid arguments 3');
            }

            $form4->getInfo($vars, $tinfo[$index]);
        }

        // Deal with attributes - FIXME: This doesn't use the normal
        // form, so if validity checking changes there, this will need
        // to be updated too.
        if (isset($cinfo['attributes']) && is_array($cinfo['attributes'])) {
            $type = $cinfo['type'];

            if (!isset($attribute_cache[$type])) {
                $attribute_cache[$type] = $whups->getAttributesForType($type);
                if (is_a($attribute_cache[$type], 'PEAR_Error')) {
                    return $attribute_cache[$type];
                }
            }

            $attributes    = $attribute_cache[$type];
            $attrib_values = array();

            foreach ($cinfo['attributes'] as $attrib) {
                $aid   = $attrib['id'];
                $value = $attrib['value'];

                if (isset($attributes[$aid])) {
                    $attrib_values["a$aid"] = $value;
                } else {
                    return PEAR::raiseError("Invalid attributes for ticket $index");
                }
            }

            if (count($attrib_values) > 0) {
                $tainfo[$index] = $attrib_values;
            }
        }
    }

    // And if they're all valid, create the tickets.
    $result = array();

    foreach ($tinfo as $index => $ticket_info) {
        $tid = $whups->addTicket($ticket_info);
        if (is_a($tid, 'PEAR_Error')) {
            $result[$index] = $tid;
        } else {
            $result[$index] = array('id' => (int)$tid, 'timestamp' => (int)time());

            if (isset($tainfo[$index])) {
                $tainfo[$index]['id'] = $tid;
                $tainfo[$index]['newcomment'] = '';

                // FIXME: need a Whups_Ticket object.
                $ticket->setAttributes($tainfo[$index]);
            }
        }
    }

    return $result;
}

function _whups_set_ticket_attributes($info)
{
    require_once dirname(__FILE__) . '/base.php';
    global $whups;

    // FIXME: Permission check on setting ticket attributes here,
    // which will also check that the ticket exists.
    if (!isset($info['ticket_id']) || !isset($info['attributes'])) {
        return PEAR::raiseError('Invalid arguments 1');
    }

    $ticket_id  = $info['ticket_id'];
    $attributes = $info['attributes'];

    if (empty($ticket_id) || empty($attributes)) {
        return PEAR::raiseError('Invalid arguments 2');
    }

    // Convert the RPC parameters into what we'd expect if we were
    // posting the EditAttributes form.
    $ainfo = array();
    foreach ($attributes as $attrib) {
        if (!isset($attrib['attribute_id']) || !isset($attrib['attribute_value'])) {
            return PEAR::raiseError('Invalid arguments 3');
        }

        $ainfo['a' . $attrib['attribute_id']] = $attrib['attribute_value'];
    }

    $ainfo['id'] = $info['ticket_id'];

    // And set 'em!
    // FIXME: need a Whups_Ticket object.
    return $ticket->setAttributes($tainfo[$index]);
}

function _whups_query($info)
{
    require_once dirname(__FILE__) . '/base.php';
    require_once dirname(__FILE__) . '/Query.php';
    global $whups;

    if (!is_array($info)) {
        return PEAR::raiseError('Invalid arguments');
    }

    foreach ($info as $index => $qinfo) {
        if (!isset($qinfo['query_name']) || !isset($qinfo['query_parameters']) || !isset($qinfo['query_details'])) {
            return PEAR::raiseError("Invalid arguments for query $index");
        }
    }

    $results = array();

    foreach ($info as $index => $qinfo) {
        $query = &new Query();
        $success = $query->load($qinfo['query_name']);

        if ($success == false)
            return PEAR::raiseError("Unknown query " . $qinfo['query_name']);

        $vars = &new Variables($qinfo['query_parameters']);
        $form = &$query->getParameterForm($vars);

        if ($form !== null) {
            $form->validate($vars);
            if (!$form->isValid())
                return PEAR::raiseError("Invalid query parameters for query $index");
        }

        $results[$index] = $whups->executeQuery($query, $vars, $qinfo['query_details'], false);

        // Now fix up the types of the results if necessary
        if ($qinfo['query_details'] === true) {
            $c = count($results[$index]);
            for ($i = 0; $i < $c; $i++) {
                settype($results[$index][$i]['id'],         'int');
                settype($results[$index][$i]['state'],      'int');
                settype($results[$index][$i]['type'],       'int');
                settype($results[$index][$i]['priority'],   'int');
                settype($results[$index][$i]['queue'],      'int');
                settype($results[$index][$i]['comment_id'], 'int');
            }
        }
    }

    return $results;
}

function _whups_getListTypes()
{
    return array('taskHash' => true);
}

function _whups_listAs($type)
{
    switch ($type) {
    case 'taskHash':
        require_once dirname(__FILE__) . '/base.php';
        global $whups;
        $info = array('owner' => 'user:' . Auth::getAuth(),
                      'nores' => true);
        $tickets = &$whups->getTicketsByProperties($info);
        if (is_a($tickets, 'PEAR_Error')) {
            return $tickets;
        }
        $result = array();
        foreach ($tickets as $ticket) {
            $view_link = Util::addParameter(Horde::applicationUrl('ticket/'), 'id', $ticket['id']);
            $delete_link = Util::addParameter(Horde::applicationUrl('ticket/delete.php'), 'id', $ticket['id']);
            $complete_link = Util::addParameter(Horde::applicationUrl('ticket/update.php'), 'id', $ticket['id']);

            $result['whups/' . $ticket['id']] = array(
                'task_id'           => $ticket['id'],
                'priority'          => $ticket['priority_name'],
                'tasklist_id'       => '**EXTERNAL**',
                'completed'         => ($ticket['state_category'] == 'resolved'),
                'name'              => $ticket['queue_name'] . ' #' .
                                       $ticket['id'] . ' - ' . $ticket['summary'],
                'desc'              => null,
                'due'               => null,
                'category'          => null,
                'view_link'         => $view_link,
                'delete_link'       => $delete_link,
                'edit_link'         => $view_link,
                'complete_link'     => $complete_link
                );
        }
        break;

    default:
        $result = array();
        break;
    }

    return $result;
}

function _whups_getQueueDetails($queue_id)
{
    require_once dirname(__FILE__) . '/base.php';
    global $whups;
    $queue = array();

    if (is_array($queue_id)) {
        foreach ($queue_id as $id) {
            $queue[] = Whups::permissionsFilter($whups->getQueue($id), 'queue', PERMS_READ);
        }
    } else {
        $queue = Whups::permissionsFilter($whups->getQueue($queue_id), 'queue', PERMS_READ);
    }

    return $queue;
}

function _whups_getTicketDetails($queue_id, $state = null)
{
    require_once dirname(__FILE__) . '/base.php';
    global $whups;

    $info['queue_id'] = $queue_id;
    if (!empty($state)) {
        $info['category'] = $state;
    }
    $tickets = $whups->getTicketsByProperties($info);

    for ($i = 0; $i < count($tickets); $i++) {
        $view_link = Util::addParameter(Horde::applicationUrl('ticket/'), 'id', $tickets[$i]['id']);
        $delete_link = Util::addParameter(Horde::applicationUrl('ticket/delete.php'), 'id', $tickets[$i]['id']);
        $complete_link = Util::addParameter(Horde::applicationUrl('ticket/update.php'), 'id', $tickets[$i]['id']);

        $tickets[$i] = array(
                'ticket_id'         => $tickets[$i]['id'],
                'completed'         => ($tickets[$i]['state_category'] == 'resolved'),
                'assigned'          => ($tickets[$i]['state_category'] == 'assigned'),
                'name'              => $tickets[$i]['queue_name'] . ' #' .
                                       $tickets[$i]['id'] . ' - ' . $tickets[$i]['summary'],
                'state'             => $tickets[$i]['state_name'],
                'type'              => $tickets[$i]['type_name'],
                'priority'          => $tickets[$i]['priority_name'],
                'desc'              => null,
                'due'               => null,
                'category'          => null,
                'view_link'         => $view_link,
                'delete_link'       => $delete_link,
                'edit_link'         => $view_link,
                'complete_link'     => $complete_link
                );
    }

    return $tickets;
}

function _whups_perms()
{
    static $perms = array();
    if (!empty($perms)) {
        return $perms;
    }

    require_once dirname(__FILE__) . '/base.php';
    global $whups;

    /* Available Whups permissions. */
    $perms['tree']['whups']['hiddenComments'] = false;
    $perms['title']['whups:hiddenComments'] = _("Hidden Comments");

    $perms['tree']['whups']['queues'] = array();
    $perms['title']['whups:queues'] = _("Queues");

    /* Loop through queues and add their titles. */
    $queues = $whups->getQueues();
    foreach ($queues as $id => $name) {
        $perms['tree']['whups']['queues'][$id] = false;
        $perms['title']['whups:queues:' . $id] = $name;
    }

    return $perms;
}
