<?php
/**
 * $Horde: whups/query.php,v 1.51 2005/01/03 14:35:43 jan Exp $
 *
 * Copyright 2001-2005 Robert E. Coyle <robertecoyle@hotmail.com>
 * Copyright 2001-2005 Chuck Hagenbuch <chuck@horde.org>
 *
 * See the enclosed file LICENSE for license information (BSD). If you
 * did not receive this file, see http://www.horde.org/licenses/bsdl.php.
 */

@define('WHUPS_BASE', dirname(__FILE__));
require_once WHUPS_BASE . '/lib/base.php';
require_once 'Horde/Variables.php';
require_once WHUPS_BASE . '/lib/Query.php';
require WHUPS_TEMPLATES . '/queryrenderer.inc';

$vars = &Variables::getDefaultVariables();

// Set up the page config vars.
$showTypeForm  = true;
$showEditQuery = true;
$showExtraForm = null;

// Query actions.
$queryurl = Horde::applicationUrl('query.php');
$QACTIONS = &new Horde_UI_Tabs('action', $vars);
$QACTIONS->addTab(_("New Query"), $queryurl, 'new');
$QACTIONS->addTab(_("Execute Query"), Horde::applicationUrl('runquery.php'), 'run');
$QACTIONS->addTab(_("Load Query"), $queryurl, 'load');
$QACTIONS->addTab(_("Save Query"), $queryurl, 'save');

// Criterion form types.
$CACTIONS = &new Horde_UI_Tabs('action', $vars);
$CACTIONS->addTab(_("Property Criteria"), $queryurl, 'props');
$CACTIONS->addTab(_("User Criteria"), $queryurl, 'user');
$CACTIONS->addTab(_("Group Criteria"), $queryurl, 'group');
// $CACTIONS->addTab(_("Date Criteria"), $queryurl, 'date');
$CACTIONS->addTab(_("Text Criteria"), $queryurl, 'text');
$CACTIONS->addTab(_("Attribute Criteria"), $queryurl, 'attribs');

$qops = array(
    ''       => _("Choose Action:"),
    'delete' => _("Delete"),
    'edit'   => _("Edit"),
    'hoist'  => _("Hoist"),
    'and'    => _("Insert And"),
    'or'     => _("Insert Or"),
    'not'    => _("Insert Not"),
);

// Find our current query.
if (isset($_SESSION['whups_query'])) {
    $whups_query = unserialize($_SESSION['whups_query']);
} else {
    $whups_query = &new Query();
}

// Find the current criteria form, and default to the user form if not
// present.
if (!isset($_SESSION['whups_query_form'])) {
    $_SESSION['whups_query_form'] = 'user';
}
$vars->set('whups_query_form', $_SESSION['whups_query_form']);

// What now? First check the result of the query edit action dropdown,
// as this action overrides the form it sits within.
if ($vars->get('qaction') != '') {
    $action = $vars->get('qaction');

    switch ($action) {
    // Query actions.
    case 'delete': $whups_query->delete($vars->get('path')); $vars->remove('path'); break;
    case 'hoist':  $whups_query->hoist($vars->get('path')); break;
    case 'branch': $form = &Horde_Form::singleton('InsertBranchForm', $vars); break;
    case 'not':    $whups_query->insertBranch($vars->get('path'), QUERY_TYPE_NOT); break;
    case 'and':    $whups_query->insertBranch($vars->get('path'), QUERY_TYPE_AND); break;
    case 'or':     $whups_query->insertBranch($vars->get('path'), QUERY_TYPE_OR); break;
    case 'edit':   echo _("Not yet implemented...");
    }

    $vars->remove('qaction');
} elseif ($vars->get('formname')) {
    // Now check for submitted forms.
    $form = &Horde_Form::singleton($vars->get('formname'), $vars);
    $form->validate($vars);

    if ($form->isValid()) {
        $form->execute($vars);
        $vars->remove('action');
    }
} elseif ($vars->get('action') != '') {
    // Last, check for actions from tabs.
    $action = $vars->get('action');

    switch ($action) {
        // Current form actions.
    case 'props':
    case 'user':
    case 'group':
    case 'date':
    case 'text':
    case 'attribs':
        $_SESSION['whups_query_form'] = $action;
        break;

        // Global query options
    case 'new':
        $whups_query->reset();
        break;

    case 'save':
        $showExtraForm = 'ChooseQueryNameForSaveForm';
        $showTypeForm  = false;
        $showEditQuery = false;
        break;

    case 'load':
        $showExtraForm = 'ChooseQueryNameForLoadForm';
        $showTypeForm  = false;
        $showEditQuery = false;
        break;
    }
}

// Start the page.
$title = _("Query Builder");
require WHUPS_TEMPLATES . '/common-header.inc';
require WHUPS_TEMPLATES . '/menu.inc';

echo $QACTIONS->render();

if ($showExtraForm !== null) {
    $form = &Horde_Form::singleton($showExtraForm, $vars);
    $form->renderActive($form->getRenderer(), $vars, 'query.php');
    echo '<br class="spacer" />';
}

/* Get the general query renderer object. */
$QUERYRENDERER = &new QueryRenderer();

if ($showEditQuery) {
    // Get our current form.
    switch ($_SESSION['whups_query_form']) {
    default:
        echo sprintf(_("Error: Unknown query form %s, defaulting to properties"), $_SESSION['whups_query_form']);
        // Fall through.

    case 'props': $form = &Horde_Form::singleton('PropertyCriterionForm', $vars); break;
    case 'user': $form = &Horde_Form::singleton('UserCriterionForm', $vars); break;
    case 'group': $form = &Horde_Form::singleton('GroupCriterionForm', $vars); break;
    case 'text': $form = &Horde_Form::singleton('TextCriterionForm', $vars); break;
    case 'attribs': $form = &Horde_Form::singleton('AttributeCriterionForm', $vars); break;
    }

    $RENDERER = &$form->getRenderer();
    $form->open($RENDERER, $vars, 'query.php', 'post');

    $QUERYRENDERER->beginActive(_("Current Query"));
    $QUERYRENDERER->edit($qops, $form->getName());
    $QUERYRENDERER->renderFormActive($whups_query, $vars);
    $RENDERER->end();

    echo '<br />';
    echo $CACTIONS->render();

    $RENDERER->beginActive($form->getTitle());
    $RENDERER->renderFormActive($form, $vars);
    $RENDERER->submit(_("Add Criterion"));
    $RENDERER->end();

    $form->close($RENDERER);

    if ($showTypeForm) {
        echo '<br class="spacer" />';
        $typeform = &Horde_Form::singleton('ChangeTypeForm', $vars);
        $typeform->renderActive($typeform->getRenderer(), $vars, 'query.php');
    }
} else {
    // Show query readonly.
    $RENDERER = &new Horde_Form_Renderer();
    $RENDERER->beginActive(_("Current Query"));
    $QUERYRENDERER->renderFormInactive($whups_query, $vars);
    $RENDERER->end();
}

require $registry->get('templates', 'horde') . '/common-footer.inc';

$_SESSION['whups_query'] = serialize($whups_query);
