<?php
/**
 * $Horde: whups/runquery.php,v 1.27 2005/01/03 14:35:43 jan Exp $
 *
 * Copyright 2001-2005 Robert E. Coyle <robertecoyle@hotmail.com>
 * Copyright 2001-2005 Chuck Hagenbuch <chuck@horde.org>
 *
 * See the enclosed file LICENSE for license information (BSD). If you
 * did not receive this file, see http://www.horde.org/licenses/bsdl.php.
 */

@define('WHUPS_BASE', dirname(__FILE__));
require_once WHUPS_BASE . '/lib/base.php';
require_once WHUPS_BASE . '/lib/Query.php';
require_once 'Horde/Variables.php';
require WHUPS_TEMPLATES . '/queryrenderer.inc';
require WHUPS_TEMPLATES . '/searchresults.inc';

$vars = &Variables::getDefaultVariables();

$queryurl = Horde::applicationUrl('query.php');
$EACTIONS = &new Horde_UI_Tabs('action', $vars);
$EACTIONS->addTab(_("Edit Query"), $queryurl, 'edit');
$EACTIONS->addTab(_("Re-execute Query"), Horde::applicationUrl('runquery.php'), 'run');
$EACTIONS->addTab(_("Load Query"), $queryurl, 'load');
$EACTIONS->addTab(_("Save Query"), $queryurl, 'save');

$RENDERER = &new Horde_Form_Renderer();
$QUERYRENDERER = &new QueryRenderer();

// Update sorting preferences.
if (Util::getFormData('sortby') !== null) {
    $prefs->setValue('sortby', Util::getFormData('sortby'));
}
if (Util::getFormData('sortdir') !== null) {
    $prefs->setValue('sortdir', Util::getFormData('sortdir'));
}

// Load the current query. If we have a 'query' parameter, that
// overrides and we load in that from the query store. Otherwise we
// use the query that is currently in our session. If we still don't
// have a query, go back to the Query Builder page.
$whups_query = null;

if ($vars->exists('query')) {
    $whups_query = &new Query();
    $whups_query->load($vars->get('query'));
} elseif (isset($_SESSION['whups_query'])) {
    $whups_query = unserialize($_SESSION['whups_query']);
} else {
    // Perhaps we're running one of a predefined set of queries and
    // we're not supposed to be able to build our own queries (site
    // admin preference).  TODO: Check for that here, and if not
    // allowed, display a "no query" message
    header('Location: ' . Horde::applicationUrl('query.php', true));
}

$form = &$whups_query->getParameterForm($vars);
$isvalid = false;
$tickets = null;

if ($form === null) {
    $isvalid = true;
} elseif ($vars->get('formname') == 'queryparameters') {
    $form->validate($vars);
    $isvalid = $form->isValid();
}

if ($isvalid) {
    $tickets = $whups->executeQuery($whups_query, $vars);
}

$title = _("Execute Query");
require WHUPS_TEMPLATES . '/common-header.inc';
require WHUPS_TEMPLATES . '/menu.inc';

echo $EACTIONS->render();

if (isset($tickets)) {
    Whups::sortTickets($tickets);
    $results = &new SearchResults(_("Query Results"));
    $results->results($tickets, Whups::getSearchResultColumns(WHUPS_RESULTS_QUERY));
} else {
    $form->open($RENDERER, $vars, 'runquery.php');
    $RENDERER->beginActive($form->getTitle());
    $RENDERER->renderFormActive($form, $vars);
    $RENDERER->submit(_("Execute Query"));
    $RENDERER->end();
    $form->close($RENDERER);
}

require $registry->get('templates', 'horde') . '/common-footer.inc';

$_SESSION['whups_query'] = serialize($whups_query);
