<?php
/**
 * $Horde: whups/search.php,v 1.80 2005/01/03 14:35:43 jan Exp $
 *
 * Copyright 2001-2005 Robert E. Coyle <robertecoyle@hotmail.com>
 * Copyright 2001-2005 Chuck Hagenbuch <chuck@horde.or>
 *
 * See the enclosed file LICENSE for license information (BSD). If you
 * did not receive this file, see http://www.horde.org/licenses/bsdl.php.
 */

@define('WHUPS_BASE', dirname(__FILE__));
require_once WHUPS_BASE . '/lib/base.php';
require_once 'Horde/Variables.php';
require_once WHUPS_TEMPLATES . '/searchresults.inc';
require_once WHUPS_TEMPLATES . '/savedsearches.inc';
require_once WHUPS_TEMPLATES . '/savedqueries.inc';

/**
 * SearchForm Class.
 *
 * @author  Robert E. Coyle <robertecoyle@hotmail.com>
 * @version $Revision: 1.80 $
 * @package Whups
 */
class SearchForm extends Horde_Form {

    var $_useFormToken = false;

    function SearchForm(&$vars)
    {
        global $whups, $conf;

        parent::Horde_Form($vars);

        $queues = Whups::permissionsFilter($whups->getQueues(), 'queue', PERMS_READ);
        if (count($queues)) {
            $modtype = 'enum';
            $type_params = array(array('0' => _("Any")) + $queues);
        } else {
            $modtype = 'invalid';
            $type_params = array(_("There are no queues which you can search."));
        }

        $count = &$this->addHidden('', 'count', 'int', false, true);
        $count->setDefault(10);

        $this->addVariable(_("Queue"), 'queue', $modtype, false, false, null, $type_params);
        $this->addVariable(_("Ticket ID"), 'id', 'intlist', false);
        $this->addVariable(_("Summary like"), 'summary', 'text', false);

        $cats = &$this->addVariable(_("Tickets which are"), 'category', 'multienum', false, false, _("If you don't select any categories, no filtering will be done on ticket category."), array($whups->getCategories(), 4));
        $cats->setDefault(array('unconfirmed', 'new', 'assigned'));

        $this->addVariable(_("Types"), 'type', 'multienum', false, false, _("If you don't select any types, no filtering will be done on ticket type."), array($whups->getAllTypes(), 4));
    }

}

$RENDERER = &new Horde_Form_Renderer();
$beendone = 0;
$vars = &Variables::getDefaultVariables();

$title = _("Search Tickets");
require WHUPS_TEMPLATES . '/common-header.inc';
require WHUPS_TEMPLATES . '/menu.inc';

$queue = $vars->getExists('queue', $wasset);
$form = &Horde_Form::singleton('SearchForm', $vars);

// Update sorting preferences.
if (Util::getFormData('sortby') !== null) {
    $prefs->setValue('sortby', Util::getFormData('sortby'));
}
if (Util::getFormData('sortdir') !== null) {
    $prefs->setValue('sortdir', Util::getFormData('sortdir'));
}

if ($wasset) {
    $form->validate($vars);

    if ($form->isValid()) {
        $form->getInfo($vars, $info);

        if (Auth::getAuth()) {
            $queueInfo = $whups->getQueue($info['queue']);
            $queueName = !empty($queueInfo['name']) ? $queueInfo['name'] : _("Any");
            $idName = $vars->get('id') ? $vars->get('id') : _("Any");
            $categories = $whups->getCategories();
            $cat = $vars->get('category');
            if ($cat) {
                if (is_array($cat)) {
                    $catName = '';
                    foreach ($cat as $c) {
                        if (!empty($catName)) {
                            $catName .= ', ';
                        }
                        $catName .= $categories[$c];
                    }
                } else {
                    $catName = $categories[$cat];
                }
            } else {
                $catName = _("Any");
            }
            $types = $whups->getAllTypes();
            $type = $vars->get('type');
            if ($type) {
                if (is_array($type)) {
                    $typeName = '';
                    foreach ($type as $t) {
                        if (!empty($typeName)) {
                            $typeName .= ', ';
                        }
                        $typeName .= $types[$t];
                    }
                } else {
                    $typeName = $types[$type];
                }
            } else {
                $typeName = _("Any");
            }

            $qUrl = Util::addParameter('', 'queue', $vars->get('queue'));
            $qUrl = Util::addParameter($qUrl, 'id', $vars->get('id'));
            if (!empty($cat)) {
                foreach ($cat as $c) {
                    $qUrl = Util::addParameter($qUrl, 'category[]', $c);
                }
            }
            if (!empty($type)) {
                foreach ($type as $t) {
                    $qUrl = Util::addParameter($qUrl, 'type[]', $t);
                }
            }
            $qUrl = substr($qUrl, 1);
            $whups->saveSearch(sprintf(_("Queue: %s, Ticket: %s, Category: %s, Type: %s"), $queueName, $idName, $catName, $typeName), $qUrl);
        }

        // Munge the search criteria into acceptability.
        if (empty($info['queue'])) {
            $info['queue'] = array_keys(Whups::permissionsFilter($whups->getQueues(), 'queue', PERMS_READ));
        }
        if (empty($info['id'])) {
            unset($info['id']);
        }
        if (empty($info['category'])) {
            unset($info['category']);
        }
        if (empty($info['type'])) {
            unset($info['type']);
        }

        $tickets = $whups->getTicketsByProperties($info);
        if (is_a($tickets, 'PEAR_Error')) {
            $notification->push(sprintf(_("There was an error performing your search: %s"), $tickets->getMessage()), 'horde.error');
        } else {
            Whups::sortTickets($tickets);

            $values = Whups::getSearchResultColumns(WHUPS_RESULTS_SEARCH);

            $results = &new SearchResults(_("Search Results"));
            $results->results($tickets, $values);
            $beendone = 1;

            if (is_object($form)) {
                $form->setTitle(_("Refine Search"));
                $form->renderActive($RENDERER, $vars, 'search.php', 'post');
                echo '<br />';

                $beendone = 1;
            }
        }
    }
}

if (!$beendone) {
    // Front search page.
    $form->setTitle(_("Search Tickets"));
    $form->renderActive($RENDERER, $vars, 'search.php', 'post');
    echo '<br />';
}

if (Auth::getAuth()) {
    $mysearches = &new SavedSearches();
    $mysearches->render(_("My QuickSearches"), $RENDERER, $whups->getUserSearches(Auth::getAuth()));

    $myqueries = &new SavedQueries();
    $myqueries->render(_("My Queries"), $RENDERER, $whups->getUserQueries(Auth::getAuth()));
}

require $registry->get('templates', 'horde') . '/common-footer.inc';
