<?php

require_once 'Horde/Text.php';

/**
 * $Horde: whups/templates/comment.inc,v 1.63 2005/01/03 14:35:45 jan Exp $
 *
 * Copyright 2001-2005 Robert E. Coyle <robertecoyle@hotmail.com>
 *
 * See the enclosed file LICENSE for license information (BSD). If you
 * did not receive this file, see http://www.horde.org/licenses/bsdl.php.
 */
class Comment extends Horde_Form_Renderer {

    var $state = null;
    var $priority = null;
    var $type = null;
    var $queue = null;

    function begin($title)
    {
        $this->_sectionHeader($title);
        echo '<br /><table cellspacing="0" width="100%">';
    }

    function render(&$vars)
    {
        global $prefs, $conf, $registry;

        $comment = '';
        $private = false;
        $changes = array();

        /* Format each change in this history entry, including
         * comments, etc. */
        $changelist = $vars->get('changes');
        foreach ($changelist as $change) {
            switch ($change['type']) {
            case 'summary':
                $changes[] = sprintf(_("Summary &rArr; %s"), htmlspecialchars($change['value']));
                break;

            case 'attachment':
                $ticket = $vars->get('ticket_id');
                $changes[] = sprintf(_("New attachment: %s"), Whups::attachmentUrl($ticket, Whups::getAttachments($ticket, $change['value'])));
                break;

            case 'assign':
                $changes[] = sprintf(_("Assigned to %s"),
                                     Whups::formatUser($change['value'], false, true, true));
                break;

            case 'unassign':
                $changes[] = sprintf(_("Taken from %s"),
                                     Whups::formatUser($change['value'], false, true, true));
                break;

            case 'comment':
                $comment = $change['comment'];
                $private = !empty($change['private']);
                break;

            case 'state':
                $changes[] = sprintf(_("State &rArr; %s"), htmlspecialchars($change['label']));
                break;

            case 'priority':
                $changes[] = sprintf(_("Priority &rArr; %s"), htmlspecialchars($change['label']));
                break;

            case 'type':
                $changes[] = sprintf(_("Type &rArr; %s"), htmlspecialchars($change['label']));
                break;

            case 'queue':
                $changes[] = sprintf(_("Queue &rArr; %s"), htmlspecialchars($change['label']));
                break;
            }
        }

        if ($comment) {
            $comment = Text_Filter::filter($comment, 'text2html', array('parselevel' => TEXT_HTML_MICRO));
            if ($prefs->getValue('autolink_tickets') && $conf['prefs']['autolink_terms']) {
                $term_regex = '/(' . $conf['prefs']['autolink_terms'] . ')\s+(\d+)/i';
                $comment = preg_replace_callback($term_regex, array(&$this, '_autolink'), $comment);
            }

            if (!count($changes)) {
                $changes[] = _("Comment");
            }
        }

        if (count($changes)) {
            ob_start();
            $class = $private ? 'pc' : 'c';
            $icon = $private ? Horde::img('locked.png', '', null, $registry->getImageDir('horde')) : '';
?>
 <tr>
  <td width="20%" class="<?php echo $class ?>_l" nowrap="nowrap" valign="top"><?php echo strftime($prefs->getValue('date_format'), $vars->get('timestamp')) ?></td>
  <td width="20%" class="<?php echo $class ?>_m" valign="top"><?php echo $vars->get('user_id') ? Whups::formatUser($vars->get('user_id'), false, true, true) : '&nbsp;' ?></td>
  <td width="60%" class="<?php echo $class ?>_r" valign="top"><?php echo implode('<br />', $changes) ?></td>
 </tr>
<?php if ($comment): ?>
 <tr><td colspan="3" class="<?php echo $class ?>_b">
  <table width="100%" cellspacing="0">
  <tr><td class="comment_body"><?php echo $icon ?></td><td width="100%" class="comment_body"><code><?php echo $comment ?></code>
  </td></tr>
  </table>
 </td></tr>
<?php else: ?>
 <tr><td colspan="3" class="c_b" style="font-size:4px;">&nbsp;</td></tr>
<?php endif; ?>
 <tr><td style="font-size:4px;">&nbsp;</td></tr>
<?php
            $html = ob_get_contents();
            ob_end_clean();
            return $html;
        }

        return '';
    }

    function _autolink($matches)
    {
        $url = Horde::applicationUrl(Util::addParameter('ticket/', 'id', $matches[2]));
        $linktext = $matches[1] . ' ' . $matches[2];
        return Horde::link($url, 'View ' . $linktext) . '<b>' . $linktext . '</b></a>';
    }

    function end()
    {
        echo '</table>';
    }

}
